import functools
import glob
import os
import sys
from typing import Any, Callable, Optional


def _memoize(f: Callable) -> Callable:
    memo = {}

    @functools.wraps(f)
    def ret(*args: Any, **kwargs: Any) -> Any:
        key = (args, frozenset(kwargs.items()))
        if key not in memo:
            memo[key] = f(*args, **kwargs)
        return memo[key]
    return ret


@_memoize
def get_nvtx_path() -> Optional[str]:
    assert sys.platform == 'win32'

    if os.environ.get('CONDA_BUILD', '0') == '1':
        return os.environ['PREFIX']

    candidates = []

    # $Env:ProgramFiles\NVIDIA Corporation\Nsight Systems 2025.3.2\
    # target-windows-x64\nvtx\include\nvtx3\nvToolsExt.h
    prog = os.environ.get('ProgramFiles', 'C:\\Program Files')
    pattern = os.path.join(
        prog, 'NVIDIA Corporation', 'Nsight Systems *', 'target-windows-x64',
        'nvtx',
    )
    candidates += sorted(p for p in glob.glob(pattern) if os.path.exists(p))

    # $Env:CUDA_PATH\include\nvtx3\nvToolsExt.h
    cuda_path = os.environ.get('CUDA_PATH', None)
    if cuda_path is not None:
        cuda_include_path = os.path.join(cuda_path, 'include')
        if os.path.exists(cuda_include_path):
            candidates += [cuda_include_path]

    print(f'Looking for NVTX from: {candidates}')
    if len(candidates) != 0:
        # Pick the one from CUDA_PATH or use the latest one from Nsight
        nvtx = candidates[-1]
        print(f'Using NVTX at: {nvtx}')
        return nvtx
    print('NVTX could not be found')
    return None



import unittest
import pytest

import numpy

import cupy
from cupy._core import core
from cupy.cuda import compiler
from cupy.cuda import runtime
from cupy import testing


def _compile_func(kernel_name, code):
    # workaround for hipRTC
    extra_source = core._get_header_source() if runtime.is_hip else None
    mod = compiler._compile_module_with_cache(
        code, options=('--std=c++11',), extra_source=extra_source)
    return mod.get_function(kernel_name)


class TestFunction(unittest.TestCase):

    def test_python_scalar(self):
        code = '''
extern "C" __global__ void test_kernel(const double* a, double b, double* x) {
  int i = threadIdx.x;
  x[i] = a[i] + b;
}
'''

        a_cpu = numpy.arange(24, dtype=numpy.float64).reshape((4, 6))
        a = cupy.array(a_cpu)
        b = float(2)
        x = cupy.empty_like(a)

        func = _compile_func('test_kernel', code)

        func.linear_launch(a.size, (a, b, x))

        expected = a_cpu + b
        testing.assert_array_equal(x, expected)

    def test_numpy_scalar(self):
        code = '''
extern "C" __global__ void test_kernel(const double* a, double b, double* x) {
  int i = threadIdx.x;
  x[i] = a[i] + b;
}
'''

        a_cpu = numpy.arange(24, dtype=numpy.float64).reshape((4, 6))
        a = cupy.array(a_cpu)
        b = numpy.float64(2)
        x = cupy.empty_like(a)

        func = _compile_func('test_kernel', code)

        func.linear_launch(a.size, (a, b, x))

        expected = a_cpu + b
        testing.assert_array_equal(x, expected)

    def test_numpy_dtype(self):
        code = '''
extern "C" __global__ void test_kernel(const double* a, double3 b, double* x) {
  int i = threadIdx.x;
  x[i] = a[i] + b.x + b.y + b.z;
}
'''

        double3 = numpy.dtype({'names': ['x', 'y', 'z'],
                               'formats': [numpy.float64]*3})
        a_cpu = numpy.arange(24, dtype=numpy.float64)
        a = cupy.array(a_cpu)
        b = numpy.random.rand(3).view(double3)
        x = cupy.empty_like(a)

        func = _compile_func('test_kernel', code)

        func.linear_launch(a.size, (a, b, x))

        expected = a_cpu + b['x'] + b['y'] + b['z']
        testing.assert_array_equal(x, expected)

    def test_static_array(self):
        code = '''
struct double5 {
    double value[5];
    __device__ const double& operator[](size_t i) const { return value[i]; }
};

extern "C" __global__ void test_kernel(const double* a, double5 b, double* x) {
  int i = threadIdx.x;
  x[i] = a[i] + b[0] + b[1] + b[2] + b[3] + b[4];
}
'''

        a_cpu = numpy.arange(24, dtype=numpy.float64)
        a = cupy.array(a_cpu)
        x = cupy.empty_like(a)

        func = _compile_func('test_kernel', code)

        # We cannot pass np.ndarray kernel arguments of size > 1
        b = numpy.arange(5).astype(numpy.float64)
        with pytest.raises(TypeError):
            func.linear_launch(a.size, (a, b, x))

        double5 = numpy.dtype({
            'names': ['dummy'],
            'formats': [(numpy.float64, (5,))]
        })
        func.linear_launch(a.size, (a, b.view(double5), x))

        expected = a_cpu + b.sum()
        testing.assert_array_equal(x, expected)

    def test_custom_user_struct(self):
        struct_definition = '''
struct custom_user_struct {
    int4 a;
    char b;
    double c[2];
    short1 d;
    unsigned long long int e[3];
};
'''

        # first step is to determine struct memory layout
        struct_layout_code = '''
{struct_definition}
extern "C" __global__ void get_struct_layout(
                                unsigned long long *itemsize,
                                unsigned long long *sizes,
                                unsigned long long *offsets) {{
    const custom_user_struct* ptr = nullptr;

    itemsize[0] = sizeof(custom_user_struct);

    sizes[0] = sizeof(ptr->a);
    sizes[1] = sizeof(ptr->b);
    sizes[2] = sizeof(ptr->c);
    sizes[3] = sizeof(ptr->d);
    sizes[4] = sizeof(ptr->e);

    offsets[0] = (unsigned long long)&ptr->a;
    offsets[1] = (unsigned long long)&ptr->b;
    offsets[2] = (unsigned long long)&ptr->c;
    offsets[3] = (unsigned long long)&ptr->d;
    offsets[4] = (unsigned long long)&ptr->e;
}}
'''.format(struct_definition=struct_definition)

        itemsize = cupy.ndarray(shape=(1,), dtype=numpy.uint64)
        sizes = cupy.ndarray(shape=(5,), dtype=numpy.uint64)
        offsets = cupy.ndarray(shape=(5,), dtype=numpy.uint64)
        func = _compile_func('get_struct_layout', struct_layout_code)
        func.linear_launch(1, (itemsize, sizes, offsets))

        # Build structure data type recursively
        names = list('abcde')
        itemsize = cupy.asnumpy(itemsize).item()
        sizes = cupy.asnumpy(sizes).tolist()
        offsets = cupy.asnumpy(offsets).tolist()

        def make_packed(basetype, N, itemsize):
            assert 0 < N <= 4, N
            names = list('xyzw')[:N]
            formats = [basetype]*N
            return numpy.dtype(dict(names=names,
                                    formats=formats,
                                    itemsize=itemsize))

        # structure member data types
        int4 = make_packed(numpy.int32, 4, sizes[0])
        char = make_packed(numpy.int8, 1, sizes[1])
        double2 = make_packed(numpy.float64, 2, sizes[2])
        short1 = make_packed(numpy.int16, 1, sizes[3])
        ulong3 = make_packed(numpy.uint64, 3, sizes[4])

        formats = [int4, char, double2, short1, ulong3]
        struct_dtype = numpy.dtype(dict(names=names,
                                        formats=formats,
                                        offsets=offsets,
                                        itemsize=itemsize))

        s = numpy.empty(shape=(1,), dtype=struct_dtype)
        s['a'] = numpy.arange(0, 4).astype(numpy.int32).view(int4)
        s['b'] = numpy.arange(4, 5).astype(numpy.int8).view(char)
        s['c'] = numpy.arange(5, 7).astype(numpy.float64).view(double2)
        s['d'] = numpy.arange(7, 8).astype(numpy.int16).view(short1)
        s['e'] = numpy.arange(8, 11).astype(numpy.uint64).view(ulong3)

        # test kernel code
        code = '''
{struct_definition}
extern "C" __global__ void test_kernel(const double* a,
                                       custom_user_struct s,
                                       double* x) {{
    int i = threadIdx.x;
    double sum = s.a.x + s.a.y + s.a.z + s.a.w;
    sum += s.b;
    sum += s.c[0] + s.c[1];
    sum += s.d.x;
    sum += s.e[0] + s.e[1] + s.e[2];
    x[i] = a[i] + sum;
}}
'''.format(struct_definition=struct_definition)

        a_cpu = numpy.arange(24, dtype=numpy.float64)
        a = cupy.array(a_cpu)
        x = cupy.empty_like(a)

        func = _compile_func('test_kernel', code)
        func.linear_launch(a.size, (a, s, x))

        expected = a_cpu + 55.0
        testing.assert_array_equal(x, expected)



cimport cpython  # NOQA

from libc cimport stdlib
from libc.stdint cimport intptr_t
from libcpp.vector cimport vector

from cupy_backends.cuda.api cimport runtime
from cupy_backends.cuda cimport stream as stream_module
from cupy._core.core cimport _ndarray_base
from cupy.cuda cimport memory

import warnings

import cupy
import cupy._core.core as core
from cupy.cuda cimport stream as py_stream_module


cdef const char* CAPSULE_NAME = "dltensor"
cdef const char* CAPSULE_NAME_VER = "dltensor_versioned"
cdef const char* USED_CAPSULE_NAME = "used_dltensor"
cdef const char* USED_CAPSULE_NAME_VER = "used_dltensor_versioned"

# The higest major and minor DLPack version currently implemented and that
# will be usually exported.
cdef uint32_t IMPL_VER_MAJOR = 1
cdef uint32_t IMPL_VER_MINOR = 0


cdef void pycapsule_deleter(object dltensor) noexcept:
    cdef DLManagedTensor* dlm_tensor
    cdef DLManagedTensorVersioned *dlm_tensor_ver

    # Do not invoke the deleter on a used capsule
    if cpython.PyCapsule_IsValid(dltensor, CAPSULE_NAME):
        dlm_tensor = <DLManagedTensor*>(
            cpython.PyCapsule_GetPointer(dltensor, CAPSULE_NAME))
        dlm_tensor.deleter(dlm_tensor)
    elif cpython.PyCapsule_IsValid(dltensor, CAPSULE_NAME_VER):
        dlm_tensor_ver = <DLManagedTensorVersioned*>(
            cpython.PyCapsule_GetPointer(dltensor, CAPSULE_NAME_VER))
        dlm_tensor_ver.deleter(dlm_tensor_ver)
    else:
        # No cleanup necessary, capsule was "consumed" (renamed).
        pass


cdef void deleter(DLManagedTensor* tensor) noexcept with gil:
    # Delete fully initialized DLManagedTensor
    stdlib.free(tensor.dl_tensor.shape)
    cpython.Py_DECREF(<_ndarray_base>tensor.manager_ctx)
    tensor.manager_ctx = NULL
    stdlib.free(tensor)


cdef void deleter_ver(DLManagedTensorVersioned* tensor) noexcept with gil:
    # Delete fully initialized DLManagedTensorVersioned
    stdlib.free(tensor.dl_tensor.shape)
    cpython.Py_DECREF(<_ndarray_base>tensor.manager_ctx)
    tensor.manager_ctx = NULL
    stdlib.free(tensor)


cdef uint8_t get_dlpack_dtype_code(dtype) except? 255:
    """Convert NumPy/CuPy to dlpack dtype (kind, without bitsize).
    """
    cdef char kind = ord(dtype.kind)

    if kind == b'u':
        return <uint8_t>kDLUInt
    elif kind == b'i':
        return <uint8_t>kDLInt
    elif kind == b'f':
        return <uint8_t>kDLFloat
    elif kind == b'c':
        return <uint8_t>kDLComplex
    elif kind == b'b':
        return <uint8_t>kDLBool
    else:
        raise BufferError('dtype is not supported for dlpack export')


cdef DLDevice get_dlpack_device(_ndarray_base array):
    cdef DLDevice device
    cdef bint is_managed

    device.device_id = array.data.device_id

    if not runtime._is_hip_environment:
        attrs = runtime.pointerGetAttributes(array.data.ptr)
        is_managed = (attrs.type == runtime.memoryTypeManaged)
        if is_managed:
            device.device_type = kDLCUDAManaged
        else:
            device.device_type = kDLCUDA
    else:
        device.device_type = kDLROCM

    return device


# The name of this function is following the framework integration guide of
# TensorComprehensions.
cpdef object toDlpack(
    _ndarray_base array, bint use_versioned=False, bint to_cpu=False,
    bint ensure_copy=False, stream=None
):
    """Create a dlpack capsule for an array.

    Parameters
    ----------
    array : ndarray
        The array to export
    use_versioned : bool
        Whether to use the versioned struct.  In the future this may be
        which version to use.
    to_cpu : bool
        Whether we should make the data CPU available.
    ensure_copy : bool
        If `to_cpu` is True, whether a copy is requested/required.
    stream : None or stream
        Only used with `to_cpu`. The stream to use for making the data
        available to the CPU.
        If `None`, we make sure to synchronize to have the data available
        as soon as we return.  Otherwise, we use this stream to copy the
        data (as requested by the user).

    .. warning::

        This function is deprecated in favor of :func:`~cupy.from_dlpack` and
        will be removed in a future version of CuPy.

    """
    warnings.warn(
        "This function is deprecated and will be removed in a future "
        "release. Use the cupy.from_dlpack() array constructor instead.",
        cupy.VisibleDeprecationWarning)
    return _toDlpack(array, use_versioned=use_versioned, to_cpu=to_cpu,
                     ensure_copy=ensure_copy, stream=stream)


cdef object _toDlpack(
    _ndarray_base array, bint use_versioned=True, bint to_cpu=False,
    bint ensure_copy=False, stream=None
):
    cdef DLManagedTensor* dlm_tensor
    cdef DLManagedTensorVersioned* dlm_tensor_ver
    cdef DLTensor* dl_tensor
    cdef const char *capsule_name

    # Fetch dtype early (as this can raise a BufferError in theory)
    cdef uint8_t dtype_code = get_dlpack_dtype_code(array.dtype)
    cdef size_t dtype_itemsize = array.dtype.itemsize

    # Fetch device information since we need it to deal with CPU logic.
    cdef DLDevice device = get_dlpack_device(array)

    cdef int32_t ndim = array._shape.size()

    if not to_cpu:
        owner = array
    elif not ensure_copy and device.device_type == kDLCUDAManaged:
        # Managed memory is CPU accessible.  Note that the consumer may expect
        # `kDLCPU` here.  We only honor this request in spirit, but not
        # strictly (because e.g. NumPy will remember the managed part).
        owner = array
        if stream is None:
            # The user did not request a stream to synchronize on.  We have to
            # assume they don't even know this is GPU data, so must fully
            # synchronize on the current stream.
            py_stream_module.get_current_stream().synchronize()
    else:
        # We need to create a CPU copy.  Assumes owner.dtype == array.dtype.
        owner = array.get(
            stream=stream, order='A', out=None, blocking=stream is None)
        device.device_type = kDLCPU
        device.device_id = 0

    cdef void *dlm_tensor_ptr = stdlib.malloc(
        sizeof(DLManagedTensorVersioned) if use_versioned
        else sizeof(DLManagedTensor)
    )
    if dlm_tensor_ptr == NULL:
        raise MemoryError()

    # Note: could coalesce this with the previous allocation in principle
    cdef int64_t* shape_strides = <int64_t*>stdlib.malloc(
        ndim * sizeof(int64_t) * 2)
    if shape_strides == NULL:
        stdlib.free(dlm_tensor_ptr)
        raise MemoryError()

    # We need a different setup for versioned/unversioned when it comes to
    # the context/deleter and additional info in the newer versioned one.
    if use_versioned:
        dlm_tensor_ver = <DLManagedTensorVersioned*>dlm_tensor_ptr

        # dl_tensor is identically filled for versioned and unversioned:
        dl_tensor = &dlm_tensor_ver.dl_tensor
        capsule_name = CAPSULE_NAME_VER

        dlm_tensor_ver.manager_ctx = <void*>owner
        dlm_tensor_ver.deleter = deleter_ver

        # "Versioned" specific initialization:
        dlm_tensor_ver.version.major = IMPL_VER_MAJOR
        dlm_tensor_ver.version.minor = IMPL_VER_MINOR

        # CuPy arrays are writeable but may be copied if copying to the CPU.
        dlm_tensor_ver.flags = 0
        if owner is not array:
            dlm_tensor_ver.flags |= DLPACK_FLAG_BITMASK_IS_COPIED
    else:
        dlm_tensor = <DLManagedTensor*>dlm_tensor_ptr
        # dl_tensor is identically filled for versioned and unversioned:
        dl_tensor = &dlm_tensor.dl_tensor
        capsule_name = CAPSULE_NAME

        dlm_tensor.manager_ctx = <void*>owner
        dlm_tensor.deleter = deleter

    # Create capsule now. After the else, the capsule will clean up on error.
    # (Note that it is good to handle expected BufferErrors early.)
    try:
        capsule = cpython.PyCapsule_New(
            dlm_tensor_ptr, capsule_name, pycapsule_deleter)
    except BaseException:
        stdlib.free(dlm_tensor_ptr)
        stdlib.free(shape_strides)
        raise
    else:
        # Finalize dlm_tensor for `pycapsule_deleter`.  This else block must
        # never fail/raise and initialize everything used by the deleter.
        cpython.Py_INCREF(owner)
        dl_tensor.shape = shape_strides

    # And fill in all other fields (that are not part of the cleanup)
    dl_tensor.ndim = ndim
    dl_tensor.strides = shape_strides + ndim
    dl_tensor.byte_offset = 0

    dl_tensor.dtype.code = dtype_code
    dl_tensor.dtype.lanes = <uint16_t>1
    dl_tensor.dtype.bits = <uint8_t>(dtype_itemsize * 8)

    dl_tensor.device = device

    # Fill in the shape and strides information (depends on GPU vs CPU array).
    # (assumes strides are a multiple of itemsize, that should be OK for cupy.)
    if owner is array:
        dl_tensor.data = <void *><intptr_t>(array.data.ptr)

        for n in range(ndim):
            shape_strides[n] = array._shape[n]

        for n in range(ndim):
            shape_strides[n + ndim] = array._strides[n] // dtype_itemsize
    else:
        # Same as above, but we got a NumPy array, so go through Python
        # in the off-chance that the copy has changed the strides.
        dl_tensor.data = <void *><intptr_t>(owner.ctypes.data)
        shape = owner.shape
        strides = owner.strides

        for n in range(ndim):
            shape_strides[n] = shape[n]

        for n in range(ndim):
            shape_strides[n + ndim] = strides[n] // dtype_itemsize

    return capsule


# TODO(leofang): Support kDLCUDAPinned and kDLROCMPinned
cdef class DLPackMemory(memory.BaseMemory):

    """Memory object for a dlpack tensor.

    This does not allocate any memory.

    """

    cdef DLManagedTensor* dlm_tensor_unversioned
    cdef DLManagedTensorVersioned* dlm_tensor_ver
    cdef DLTensor* dl_tensor  # non owning reference

    def __init__(self, object dltensor):
        # First, take ownership of the contained DLManagedTensor(Versioned)
        # by copying it over and setting the capsule name to "used".
        if cpython.PyCapsule_IsValid(dltensor, CAPSULE_NAME):
            # Take ownership of the memory:
            self.dlm_tensor_unversioned = <DLManagedTensor*>(
                cpython.PyCapsule_GetPointer(dltensor, CAPSULE_NAME))
            cpython.PyCapsule_SetName(dltensor, USED_CAPSULE_NAME)

            self.dl_tensor = &self.dlm_tensor_unversioned.dl_tensor
        elif cpython.PyCapsule_IsValid(dltensor, CAPSULE_NAME_VER):
            # Take ownership of the memory:
            self.dlm_tensor_ver = <DLManagedTensorVersioned*>(
                cpython.PyCapsule_GetPointer(dltensor, CAPSULE_NAME_VER))
            cpython.PyCapsule_SetName(dltensor, USED_CAPSULE_NAME_VER)

            # When we have a versioned tensor, we need to verify the version
            # before further use.
            if self.dlm_tensor_ver.version.major > IMPL_VER_MAJOR:
                raise BufferError("DLPack exported too new major version.")

            # TODO(seberg): In principle we should raise an error if we got a
            #     readonly buffer.  But that might be disruptive :(.
            # if self.dlm_tensor_ver.flags & DLPACK_FLAG_BITMASK_READ_ONLY:
            #  raise BufferError("Buffer is readonly, but...")

            self.dl_tensor = &self.dlm_tensor_ver.dl_tensor
        else:
            # Be helpful in case a capsule is used twice somehow:
            raise ValueError(
                'A DLPack tensor object cannot be consumed multiple times '
                f'(or object was not a DLPack capsule). Got: {dltensor!r}')

        # Check if the device is compatible with the cupy runtime.
        if runtime._is_hip_environment:
            if self.dl_tensor.device.device_type != kDLROCM:
                raise RuntimeError('CuPy is built against ROCm/HIP, different '
                                   'from the backend that backs the incoming '
                                   'DLPack tensor')
        else:
            if self.dl_tensor.device.device_type not in (
                    kDLCUDA, kDLCUDAManaged):
                raise RuntimeError('CuPy is built against CUDA, different '
                                   'from the backend that backs the incoming '
                                   'DLPack tensor')

        self.ptr = <intptr_t>self.dl_tensor.data

        if self.dl_tensor.device.device_type == kDLCUDAManaged:
            # look up the actual physical device as the id from
            # dl_tensor could be 0
            attrs = runtime.pointerGetAttributes(self.ptr)
            self.device_id = attrs.device
        else:
            self.device_id = self.dl_tensor.device.device_id

        cdef int n = 0, s = 0
        cdef int ndim = self.dl_tensor.ndim
        cdef int64_t* shape = self.dl_tensor.shape
        for s in shape[:ndim]:
            n += s
        self.size = self.dl_tensor.dtype.bits * n // 8

    def __dealloc__(self):
        # dlm_tensor could be uninitialized if an error is raised in __init__
        if self.dlm_tensor_unversioned != NULL:
            if self.dlm_tensor_unversioned.deleter != NULL:
                self.dlm_tensor_unversioned.deleter(
                    self.dlm_tensor_unversioned)
        elif self.dlm_tensor_ver != NULL:
            if self.dlm_tensor_ver.deleter != NULL:
                self.dlm_tensor_ver.deleter(self.dlm_tensor_ver)


# The name of this function is following the framework integration guide of
# TensorComprehensions.
cpdef _ndarray_base fromDlpack(object dltensor):
    """Zero-copy conversion from a DLPack tensor to a :class:`~cupy.ndarray`.

    DLPack is a open in memory tensor structure proposed in this repository:
    `dmlc/dlpack <https://github.com/dmlc/dlpack>`_.

    This function takes a :class:`PyCapsule` object which contains a pointer to
    a DLPack tensor as input, and returns a :class:`~cupy.ndarray`. This
    function does not copy the data in the DLPack tensor but both
    DLPack tensor and :class:`~cupy.ndarray` have pointers which are pointing
    to the same memory region for the data.

    Args:
        dltensor (:class:`PyCapsule`): Input DLPack tensor which is
            encapsulated in a :class:`PyCapsule` object.

    Returns:
        array (:class:`~cupy.ndarray`): A CuPy ndarray.

    .. warning::

        This function is deprecated in favor of :func:`~cupy.from_dlpack` and
        will be removed in a future version of CuPy.

    .. warning::

        As of the DLPack v0.5 specification, it is implicitly assumed that
        the user is responsible to ensure the Producer and the Consumer are
        operating on the same stream.

    .. seealso::

        :meth:`cupy.ndarray.toDlpack` is a method for zero-copy conversion
        from a :class:`~cupy.ndarray` to a DLPack tensor (which is encapsulated
        in a :class:`PyCapsule` object).

    .. admonition:: Example

        >>> import cupy
        >>> array1 = cupy.array([0, 1, 2], dtype=cupy.float32)
        >>> dltensor = array1.toDlpack()
        >>> array2 = cupy.fromDlpack(dltensor)
        >>> cupy.testing.assert_array_equal(array1, array2)

    """
    warnings.warn(
        "This function is deprecated and will be removed in a future "
        "release. Use the cupy.from_dlpack() array constructor instead.",
        cupy.VisibleDeprecationWarning)
    return _dlpack_to_cupy_array(dltensor)


cdef inline _ndarray_base _dlpack_to_cupy_array(dltensor):
    cdef DLPackMemory mem = DLPackMemory(dltensor)
    cdef DLDataType dtype = mem.dl_tensor.dtype
    cdef int bits = dtype.bits
    if dtype.lanes != 1:
        raise ValueError(f'vector dtypes (lanes={dtype.lanes}) is '
                         'not supported')
    if dtype.code == kDLUInt:
        if bits == 8:
            cp_dtype = cupy.uint8
        elif bits == 16:
            cp_dtype = cupy.uint16
        elif bits == 32:
            cp_dtype = cupy.uint32
        elif bits == 64:
            cp_dtype = cupy.uint64
        else:
            raise TypeError('uint{} is not supported.'.format(bits))
    elif dtype.code == kDLInt:
        if bits == 8:
            cp_dtype = cupy.int8
        elif bits == 16:
            cp_dtype = cupy.int16
        elif bits == 32:
            cp_dtype = cupy.int32
        elif bits == 64:
            cp_dtype = cupy.int64
        else:
            raise TypeError('int{} is not supported.'.format(bits))
    elif dtype.code == kDLFloat:
        if bits == 16:
            cp_dtype = cupy.float16
        elif bits == 32:
            cp_dtype = cupy.float32
        elif bits == 64:
            cp_dtype = cupy.float64
        else:
            raise TypeError('float{} is not supported.'.format(bits))
    elif dtype.code == kDLComplex:
        # TODO(leofang): support complex32
        if bits == 64:
            cp_dtype = cupy.complex64
        elif bits == 128:
            cp_dtype = cupy.complex128
        else:
            raise TypeError('complex{} is not supported.'.format(bits))
    elif dtype.code == kDLBool:
        if bits == 8:
            cp_dtype = cupy.bool_
        else:
            raise TypeError(f'{bits}-bit bool is not supported')
    elif dtype.code == kDLBfloat:
        raise NotImplementedError('CuPy does not support bfloat16 yet')
    else:
        raise TypeError('Unsupported dtype. dtype code: {}'.format(dtype.code))

    mem_ptr = memory.MemoryPointer(mem, mem.dl_tensor.byte_offset)
    cdef int64_t ndim = mem.dl_tensor.ndim

    cdef int64_t* shape = mem.dl_tensor.shape
    cdef vector[Py_ssize_t] shape_vec
    shape_vec.assign(shape, shape + ndim)

    if mem.dl_tensor.strides is NULL:
        return core.ndarray(shape_vec, cp_dtype, mem_ptr, strides=None)
    cdef int64_t* strides = mem.dl_tensor.strides
    cdef vector[Py_ssize_t] strides_vec
    for i in range(ndim):
        strides_vec.push_back(strides[i] * (bits // 8))

    return core.ndarray(shape_vec, cp_dtype, mem_ptr, strides=strides_vec)


def from_dlpack(array, *, device=None, copy=None):
    """Zero-copy conversion between array objects compliant with the DLPack
    data exchange protocol.

    Args:
        array (object): an array object that implements two methods:
            ``__dlpack__()`` and ``__dlpack_device__()``.
        device (tuple): The dlpack device as a ``(device_type, device_id)``
            tuple.
        copy (boolean|None): Request export to never or always make a copy.
            By default (``None``) the exporter may or may not make a copy.

    Returns:
        cupy.ndarray: a CuPy array that can be safely accessed on CuPy's
        current stream.

    .. note::
        This function is different from CuPy's legacy :func:`~cupy.fromDlpack`
        function. This function takes any object implementing the DLPack data
        exchange protocol, as well as a raw :class:`PyCapsule` object that
        contains the DLPack tensor as input (for backward compatibility),
        whereas :func:`~cupy.fromDlpack` only accepts :class:`PyCapsule`
        objects. If the input object is not compliant with the protocol, users
        are responsible to ensure data safety.

    .. seealso::
        :func:`numpy.from_dlpack`,
        `Python Specification for DLPack`_,
        `Data interchange mechanisms`_

    .. _Python Specification for DLPack:
        https://dmlc.github.io/dlpack/latest/python_spec.html
    .. _Data interchange mechanisms:
        https://data-apis.org/array-api/latest/design_topics/data_interchange.html
    """
    if not hasattr(array, '__dlpack_device__'):
        # backward compatibility: accept passing in a pycapsule
        dltensor = array
        return _dlpack_to_cupy_array(dltensor)
    else:
        dev_type, dev_id = array.__dlpack_device__()

    if device is not None:
        # We can probably support the user picking a specific GPU in case that
        # is relevant to them for some reason.
        raise NotImplementedError("from_dlpack() does not support device yet.")

    # CuPy is the consumer, so we provide our current stream to the producer
    if dev_type == <int>kDLCUDA or dev_type == <int>kDLCUDAManaged:
        prev_device = cupy.cuda.runtime.getDevice()
        try:
            cupy.cuda.runtime.setDevice(dev_id)
            assert not runtime._is_hip_environment
            stream = stream_module.get_current_stream_ptr()
            if stream == 0:
                stream = stream_module.get_default_stream_ptr()

            # For backwards compatibility we catch TypeErrors for now.
            try:
                dltensor = array.__dlpack__(
                    stream=stream,
                    max_version=(IMPL_VER_MAJOR, IMPL_VER_MINOR),
                    copy=copy
                )
            except TypeError:
                if copy is not None:
                    raise
                dltensor = array.__dlpack__(stream=stream)
        finally:
            cupy.cuda.runtime.setDevice(prev_device)
    elif dev_type == <int>kDLROCM:
        prev_device = cupy.cuda.runtime.getDevice()
        try:
            cupy.cuda.runtime.setDevice(dev_id)
            assert runtime._is_hip_environment
            stream = stream_module.get_current_stream_ptr()

            # For backwards compatibility we catch TypeErrors for now.
            try:
                dltensor = array.__dlpack__(
                    stream=stream,
                    max_version=(IMPL_VER_MAJOR, IMPL_VER_MINOR),
                    copy=copy
                )
            except TypeError:
                if copy is not None:
                    raise
                dltensor = array.__dlpack__(stream=stream)
        finally:
            cupy.cuda.runtime.setDevice(prev_device)
    elif dev_type == <int>kDLCPU:
        raise TypeError(
            'CPU arrays cannot be directly imported to CuPy. '
            'Use `cupy.array(numpy.from_dlpack(input))` instead.')
    else:
        # TODO(leofang): support kDLCUDAPinned etc
        raise TypeError(f'Unsupported array type: {dev_type}')

    return _dlpack_to_cupy_array(dltensor)



"""Beta and log(abs(beta)) functions.

Also the incomplete beta function and its inverse.

The source code here is an adaptation with minimal changes from the following
files in SciPy's bundled Cephes library:

https://github.com/scipy/scipy/blob/main/scipy/special/cephes/beta.c
https://github.com/scipy/scipy/blob/main/scipy/special/cephes/incbet.c
https://github.com/scipy/scipy/blob/main/scipy/special/cephes/incbi.c

Cephes Math Library, Release 2.3:  March, 1995
Copyright 1984, 1995 by Stephen L. Moshier
"""

from cupy import _core
from cupyx.scipy.special._digamma import polevl_definition
from cupyx.scipy.special._gamma import gamma_definition
from cupyx.scipy.special._gammainc import p1evl_definition


beta_preamble = """
// include for CUDART_INF, CUDART_NAN
#include <cupy/math_constants.h>

// defines from /scipy/special/cephes/const.c
// The ones chosen here are the IEEE variants from that file
__constant__ double MAXLOG = 7.09782712893383996732E2;
__constant__ double MINLOG = -7.08396418532264106224E2;
__constant__ double MACHEP = 1.11022302462515654042E-16;  // 2**-53

// defines from npy_math.h
#define NPY_PI        3.141592653589793238462643383279502884  /* pi */

#define MAXGAM 171.624376956302725
#define ASYMP_FACTOR 1e6
"""


lgam_sgn_definition = """


/* A[]: Stirling's formula expansion of log Gamma
 * B[], C[]: log Gamma function between 2 and 3
 */
__constant__ double A[] = {
    8.11614167470508450300E-4,
    -5.95061904284301438324E-4,
    7.93650340457716943945E-4,
    -2.77777777730099687205E-3,
    8.33333333333331927722E-2
};

__constant__ double B[] = {
    -1.37825152569120859100E3,
    -3.88016315134637840924E4,
    -3.31612992738871184744E5,
    -1.16237097492762307383E6,
    -1.72173700820839662146E6,
    -8.53555664245765465627E5
};

__constant__ double C[] = {
    /* 1.00000000000000000000E0, */
    -3.51815701436523470549E2,
    -1.70642106651881159223E4,
    -2.20528590553854454839E5,
    -1.13933444367982507207E6,
    -2.53252307177582951285E6,
    -2.01889141433532773231E6
};

/* log( sqrt( 2*pi ) ) */
__constant__ double LS2PI = 0.91893853320467274178;

__constant__ double LOGPI = 1.14472988584940017414;

#define MAXLGM 2.556348e305


__noinline__ __device__ static double lgam_sgn(double x, int *sign)
{
    double p, q, u, w, z;
    int i;

    *sign = 1;

    if (isinf(x)) {
        return x;
    }

    if (x < -34.0) {
        q = -x;
        w = lgam_sgn(q, sign);
        p = floor(q);
        if (p == q) {
            return CUDART_INF;
        }
        i = p;
        if ((i & 1) == 0)
            *sign = -1;
        else
            *sign = 1;
        z = q - p;
        if (z > 0.5) {
            p += 1.0;
            z = p - q;
        }
        z = q * sin(NPY_PI * z);
        if (z == 0.0){
            return CUDART_INF;
        }
        /*     z = log(NPY_PI) - log( z ) - w; */
        z = LOGPI - log(z) - w;
        return z;
    }

    if (x < 13.0) {
        z = 1.0;
        p = 0.0;
        u = x;
        while (u >= 3.0) {
            p -= 1.0;
            u = x + p;
            z *= u;
        }
        while (u < 2.0) {
            if (u == 0.0) {
                return CUDART_INF;
            }
            z /= u;
            p += 1.0;
            u = x + p;
        }
        if (z < 0.0) {
            *sign = -1;
            z = -z;
        }
        else {
            *sign = 1;
        }
        if (u == 2.0) {
            return log(z);
        }
        p -= 2.0;
        x = x + p;
        p = x * polevl<5>(x, B) / p1evl<6>(x, C);
        return (log(z) + p);
    }

    if (x > MAXLGM) {
        return (*sign * CUDART_INF);
    }

    q = (x - 0.5) * log(x) - x + LS2PI;
    if (x > 1.0e8) {
        return q;
    }

    p = 1.0 / (x * x);
    if (x >= 1000.0) {
        q += ((7.9365079365079365079365e-4 * p
               - 2.7777777777777777777778e-3) * p
              + 0.0833333333333333333333) / x;
    } else {
        q += polevl<4>(p, A) / x;
    }
    return q;
}


__device__ static double lgam(double x)
{
    int sign;
    return lgam_sgn(x, &sign);
}

"""

lbeta_symp_definition = """
/*
 * Asymptotic expansion for  ln(|B(a, b)|) for a > ASYMP_FACTOR*max(|b|, 1).
 */
__noinline__ __device__ double lbeta_asymp(double a, double b, int *sgn)
{
    double r = lgam_sgn(b, sgn);
    r -= b * log(a);

    r += b*(1-b)/(2*a);
    r += b*(1-b)*(1-2*b)/(12*a*a);
    r += - b*b*(1-b)*(1-b)/(12*a*a*a);

    return r;
}
"""


beta_definition = """

__noinline__ __device__ double beta(double, double);


/*
 * Special case for a negative integer argument
 */

__noinline__ __device__ static double beta_negint(int a, double b)
{
    int sgn;
    if (b == (int)b && 1 - a - b > 0) {
        sgn = ((int)b % 2 == 0) ? 1 : -1;
        return sgn * beta(1 - a - b, b);
    }
    else {
        return CUDART_INF;
    }
}

__noinline__ __device__ double beta(double a, double b)
{
    double y;
    int sign = 1;

    if (a <= 0.0) {
        if (a == floor(a)) {
            if (a == (int)a) {
                return beta_negint((int)a, b);
            }
            else {
                return CUDART_INF;
            }
        }
    }

    if (b <= 0.0) {
        if (b == floor(b)) {
            if (b == (int)b) {
                return beta_negint((int)b, a);
            }
            else {
                return CUDART_INF;
            }
        }
    }

    if (fabs(a) < fabs(b)) {
        y = a; a = b; b = y;
    }

    if (fabs(a) > ASYMP_FACTOR * fabs(b) && a > ASYMP_FACTOR) {
        /* Avoid loss of precision in lgam(a + b) - lgam(a) */
        y = lbeta_asymp(a, b, &sign);
        return sign * exp(y);
    }

    y = a + b;
    if (fabs(y) > MAXGAM || fabs(a) > MAXGAM || fabs(b) > MAXGAM) {
        int sgngam;
        y = lgam_sgn(y, &sgngam);
        sign *= sgngam;     /* keep track of the sign */
        y = lgam_sgn(b, &sgngam) - y;
        sign *= sgngam;
        y = lgam_sgn(a, &sgngam) + y;
        sign *= sgngam;
        if (y > MAXLOG) {
            return sign * CUDART_INF;
        }
        return sign * exp(y);
    }

    y = Gamma(y);
    a = Gamma(a);
    b = Gamma(b);
    if (y == 0.0) {
        return sign * CUDART_INF;
    }

    if (fabs(fabs(a) - fabs(y)) > fabs(fabs(b) - fabs(y))) {
        y = b / y;
        y *= a;
    }
    else {
        y = a / y;
        y *= b;
    }

    return (y);
}
"""


lbeta_definition = """

__noinline__ __device__ double lbeta(double, double);


/*
 * Special case for a negative integer argument
 */

__noinline__ __device__ static double lbeta_negint(int a, double b)
{
    double r;
    if (b == (int)b && 1 - a - b > 0) {
        r = lbeta(1 - a - b, b);
        return r;
    }
    else {
        return CUDART_INF;
    }
}

// Natural log of |beta|

__noinline__ __device__ double lbeta(double a, double b)
{
    double y;
    int sign;

    sign = 1;

    if (a <= 0.0) {
        if (a == floor(a)) {
            if (a == (int)a) {
                return lbeta_negint((int)a, b);
            }
            else {
                return CUDART_INF;
            }
        }
    }

    if (b <= 0.0) {
        if (b == floor(b)) {
            if (b == (int)b) {
                return lbeta_negint((int)b, a);
            }
            else {
                return CUDART_INF;
            }
        }
    }

    if (fabs(a) < fabs(b)) {
        y = a; a = b; b = y;
    }

    if (fabs(a) > ASYMP_FACTOR * fabs(b) && a > ASYMP_FACTOR) {
        /* Avoid loss of precision in lgam(a + b) - lgam(a) */
        y = lbeta_asymp(a, b, &sign);
        return y;
    }

    y = a + b;
    if (fabs(y) > MAXGAM || fabs(a) > MAXGAM || fabs(b) > MAXGAM) {
        int sgngam;
        y = lgam_sgn(y, &sgngam);
        sign *= sgngam;     /* keep track of the sign */
        y = lgam_sgn(b, &sgngam) - y;
        sign *= sgngam;
        y = lgam_sgn(a, &sgngam) + y;
        sign *= sgngam;
        return (y);
    }

    y = Gamma(y);
    a = Gamma(a);
    b = Gamma(b);
    if (y == 0.0) {
        return (sign * CUDART_INF);
    }

    if (fabs(fabs(a) - fabs(y)) > fabs(fabs(b) - fabs(y))) {
        y = b / y;
        y *= a;
    }
    else {
        y = a / y;
        y *= b;
    }

    if (y < 0) {
        y = -y;
    }
    return log(y);
}

"""


beta = _core.create_ufunc(
    "cupyx_scipy_beta",
    ("ff->f", "dd->d"),
    "out0 = out0_type(beta(in0, in1));",
    preamble=(
        beta_preamble +
        gamma_definition +
        polevl_definition +
        p1evl_definition +
        lgam_sgn_definition +
        lbeta_symp_definition +
        beta_definition
    ),
    doc="""Beta function.

    Parameters
    ----------
    a, b : cupy.ndarray
        Real-valued arguments
    out : cupy.ndarray, optional
        Optional output array for the function result

    Returns
    -------
    scalar or ndarray
        Value of the beta function

    See Also
    --------
    :func:`scipy.special.beta`

    """,
)


betaln = _core.create_ufunc(
    "cupyx_scipy_betaln",
    ("ff->f", "dd->d"),
    "out0 = out0_type(lbeta(in0, in1));",
    preamble=(
        beta_preamble +
        gamma_definition +
        polevl_definition +
        p1evl_definition +
        lgam_sgn_definition +
        lbeta_symp_definition +
        lbeta_definition
    ),
    doc="""Natural logarithm of absolute value of beta function.

    Computes ``ln(abs(beta(a, b)))``.

    Parameters
    ----------
    a, b : cupy.ndarray
        Real-valued arguments
    out : cupy.ndarray, optional
        Optional output array for the function result

    Returns
    -------
    scalar or ndarray
        Value of the natural log of the magnitude of beta.

    See Also
    --------
    :func:`scipy.special.betaln`

    """,
)


incbet_definition = """

__noinline__ __device__ static double incbd(double, double, double);
__noinline__ __device__ static double incbcf(double, double, double);
__noinline__ __device__ static double pseries(double, double, double);

__constant__ double big = 4.503599627370496e15;
__constant__ double biginv = 2.22044604925031308085e-16;

__noinline__ __device__ double incbet(double aa, double bb, double xx)
{
    double a, b, t, x, xc, w, y;
    int flag;

    if (aa <= 0.0 || bb <= 0.0)
    {
        return CUDART_NAN;
    }

    if ((xx <= 0.0) || (xx >= 1.0)) {
        if (xx == 0.0) {
            return 0.0;
        }
        if (xx == 1.0) {
            return 1.0;
        }
        return CUDART_NAN;
    }

    flag = 0;
    if ((bb * xx) <= 1.0 && xx <= 0.95) {
        t = pseries(aa, bb, xx);
        goto done;
    }

    w = 1.0 - xx;

    /* Reverse a and b if x is greater than the mean. */
    if (xx > (aa / (aa + bb))) {
        flag = 1;
        a = bb;
        b = aa;
        xc = xx;
        x = w;
    } else {
        a = aa;
        b = bb;
        xc = w;
        x = xx;
    }

    if (flag == 1 && (b * x) <= 1.0 && x <= 0.95) {
        t = pseries(a, b, x);
        goto done;
    }

    /* Choose expansion for better convergence. */
    y = x * (a + b - 2.0) - (a - 1.0);
    if (y < 0.0) {
        w = incbcf(a, b, x);
    } else {
        w = incbd(a, b, x) / xc;
    }

    /* Multiply w by the factor
     * a      b   _             _     _
     * x  (1-x)   | (a+b) / ( a | (a) | (b) ) .   */

    y = a * log(x);
    t = b * log(xc);
    if ((a + b) < MAXGAM && fabs(y) < MAXLOG && fabs(t) < MAXLOG) {
        t = pow(xc, b);
        t *= pow(x, a);
        t /= a;
        t *= w;
        t *= 1.0 / beta(a, b);
        goto done;
    }
    /* Resort to logarithms.  */
    y += t - lbeta(a,b);
    y += log(w / a);
    if (y < MINLOG) {
        t = 0.0;
    } else {
        t = exp(y);
    }

  done:

    if (flag == 1) {
        if (t <= MACHEP) {
            t = 1.0 - MACHEP;
        } else {
            t = 1.0 - t;
        }
    }
    return t;
}


/* Continued fraction expansion #1
 * for incomplete beta integral
 */

__noinline__ __device__ static double incbcf(double a, double b, double x)
{
    double xk, pk, pkm1, pkm2, qk, qkm1, qkm2;
    double k1, k2, k3, k4, k5, k6, k7, k8;
    double r, t, ans, thresh;
    int n;

    k1 = a;
    k2 = a + b;
    k3 = a;
    k4 = a + 1.0;
    k5 = 1.0;
    k6 = b - 1.0;
    k7 = k4;
    k8 = a + 2.0;

    pkm2 = 0.0;
    qkm2 = 1.0;
    pkm1 = 1.0;
    qkm1 = 1.0;
    ans = 1.0;
    r = 1.0;
    n = 0;
    thresh = 3.0 * MACHEP;
    do {

        xk = -(x * k1 * k2) / (k3 * k4);
        pk = pkm1 + pkm2 * xk;
        qk = qkm1 + qkm2 * xk;
        pkm2 = pkm1;
        pkm1 = pk;
        qkm2 = qkm1;
        qkm1 = qk;

        xk = (x * k5 * k6) / (k7 * k8);
        pk = pkm1 + pkm2 * xk;
        qk = qkm1 + qkm2 * xk;
        pkm2 = pkm1;
        pkm1 = pk;
        qkm2 = qkm1;
        qkm1 = qk;

        if (qk != 0)
            r = pk / qk;
        if (r != 0) {
            t = fabs((ans - r) / r);
            ans = r;
        } else {
            t = 1.0;
        }

        if (t < thresh) {
            return ans;
        }

        k1 += 1.0;
        k2 += 1.0;
        k3 += 2.0;
        k4 += 2.0;
        k5 += 1.0;
        k6 -= 1.0;
        k7 += 2.0;
        k8 += 2.0;

        if ((fabs(qk) + fabs(pk)) > big) {
            pkm2 *= biginv;
            pkm1 *= biginv;
            qkm2 *= biginv;
            qkm1 *= biginv;
        }
        if ((fabs(qk) < biginv) || (fabs(pk) < biginv)) {
            pkm2 *= big;
            pkm1 *= big;
            qkm2 *= big;
            qkm1 *= big;
        }
    }
    while (++n < 300);

    return ans;
}



/* Continued fraction expansion #2
 * for incomplete beta integral
 */

__noinline__ __device__ static double incbd(double a, double b, double x)
{
    double xk, pk, pkm1, pkm2, qk, qkm1, qkm2;
    double k1, k2, k3, k4, k5, k6, k7, k8;
    double r, t, ans, z, thresh;
    int n;

    k1 = a;
    k2 = b - 1.0;
    k3 = a;
    k4 = a + 1.0;
    k5 = 1.0;
    k6 = a + b;
    k7 = a + 1.0;;
    k8 = a + 2.0;

    pkm2 = 0.0;
    qkm2 = 1.0;
    pkm1 = 1.0;
    qkm1 = 1.0;
    z = x / (1.0 - x);
    ans = 1.0;
    r = 1.0;
    n = 0;
    thresh = 3.0 * MACHEP;
    do {

        xk = -(z * k1 * k2) / (k3 * k4);
        pk = pkm1 + pkm2 * xk;
        qk = qkm1 + qkm2 * xk;
        pkm2 = pkm1;
        pkm1 = pk;
        qkm2 = qkm1;
        qkm1 = qk;

        xk = (z * k5 * k6) / (k7 * k8);
        pk = pkm1 + pkm2 * xk;
        qk = qkm1 + qkm2 * xk;
        pkm2 = pkm1;
        pkm1 = pk;
        qkm2 = qkm1;
        qkm1 = qk;

        if (qk != 0){
            r = pk / qk;
        }
        if (r != 0) {
            t = fabs((ans - r) / r);
            ans = r;
        } else {
            t = 1.0;
        }

        if (t < thresh) {
            return ans;
        }

        k1 += 1.0;
        k2 -= 1.0;
        k3 += 2.0;
        k4 += 2.0;
        k5 += 1.0;
        k6 += 1.0;
        k7 += 2.0;
        k8 += 2.0;

        if ((fabs(qk) + fabs(pk)) > big) {
            pkm2 *= biginv;
            pkm1 *= biginv;
            qkm2 *= biginv;
            qkm1 *= biginv;
        }
        if ((fabs(qk) < biginv) || (fabs(pk) < biginv)) {
            pkm2 *= big;
            pkm1 *= big;
            qkm2 *= big;
            qkm1 *= big;
        }
    }
    while (++n < 300);

    return ans;
}


/* Power series for incomplete beta integral.
 * Use when b*x is small and x not too close to 1.  */

__noinline__ __device__ static double pseries(double a, double b, double x)
{
    double s, t, u, v, n, t1, z, ai;

    ai = 1.0 / a;
    u = (1.0 - b) * x;
    v = u / (a + 1.0);
    t1 = v;
    t = u;
    n = 2.0;
    s = 0.0;
    z = MACHEP * ai;
    while (fabs(v) > z) {
        u = (n - b) * x / n;
        t *= u;
        v = t / (a + n);
        s += v;
        n += 1.0;
    }
    s += t1;
    s += ai;

    u = a * log(x);
    if ((a + b) < MAXGAM && fabs(u) < MAXLOG) {
        t = 1.0 / beta(a, b);
    s = s * t * pow(x, a);
    } else {
        t = -lbeta(a,b) + u + log(s);
        if (t < MINLOG) {
            s = 0.0;
        } else {
            s = exp(t);
        }
    }
    return (s);
}


"""

incbet_preamble = (
    beta_preamble +
    gamma_definition +
    polevl_definition +
    p1evl_definition +
    lgam_sgn_definition +
    lbeta_symp_definition +
    beta_definition +
    lbeta_definition +
    incbet_definition
)


betainc = _core.create_ufunc(
    "cupyx_scipy_betainc",
    ("fff->f", "ddd->d"),
    "out0 = out0_type(incbet(in0, in1, in2));",
    preamble=incbet_preamble,
    doc="""Incomplete beta function.

    Parameters
    ----------
    a, b : cupy.ndarray
        Positive, real-valued parameters
    x : cupy.ndarray
        Real-valued such that 0 <= x <= 1, the upper limit of integration.
    out : ndarray, optional
        Optional output array for the function result

    Returns
    -------
    scalar or ndarray
        Value of the incomplete beta function

    See Also
    --------
    :func:`scipy.special.betainc`

    """,
)


incbi_definition = """

__noinline__ __device__ double incbi(double aa, double bb, double yy0)
{
    double a, b, y0, d, y, x, x0, x1, lgm, yp, di, dithresh, yl, yh, xt;
    int i, rflg, dir, nflg;

    i = 0;
    if (yy0 <= 0) {
        return 0.0;
    }
    if (yy0 >= 1.0) {
        return 1.0;
    }
    x0 = 0.0;
    yl = 0.0;
    x1 = 1.0;
    yh = 1.0;
    nflg = 0;

    if (aa <= 1.0 || bb <= 1.0) {
        dithresh = 1.0e-6;
        rflg = 0;
        a = aa;
        b = bb;
        y0 = yy0;
        x = a / (a + b);
        y = incbet(a, b, x);
        goto ihalve;
    }
    else {
        dithresh = 1.0e-4;
    }
    /* approximation to inverse function */

    // normcdfinv is the CUDA Math API equivalent of cephes ndtri
    yp = -normcdfinv(yy0);

    if (yy0 > 0.5) {
        rflg = 1;
        a = bb;
        b = aa;
        y0 = 1.0 - yy0;
        yp = -yp;
    }
    else {
        rflg = 0;
        a = aa;
        b = bb;
        y0 = yy0;
    }

    lgm = (yp * yp - 3.0) / 6.0;
    x = 2.0 / (1.0 / (2.0 * a - 1.0) + 1.0 / (2.0 * b - 1.0));
    d = yp * sqrt(x + lgm) / x
        - (1.0 / (2.0 * b - 1.0) - 1.0 / (2.0 * a - 1.0))
        * (lgm + 5.0 / 6.0 - 2.0 / (3.0 * x));
    d = 2.0 * d;
    if (d < MINLOG) {
        x = 1.0;
        goto under;
    }
    x = a / (a + b * exp(d));
    y = incbet(a, b, x);
    yp = (y - y0) / y0;
    if (fabs(yp) < 0.2) {
        goto newt;
    }

    /* Resort to interval halving if not close enough. */
ihalve:

    dir = 0;
    di = 0.5;
    for (i = 0; i < 100; i++) {
        if (i != 0) {
            x = x0 + di * (x1 - x0);
            if (x == 1.0) {
                x = 1.0 - MACHEP;
            }
            if (x == 0.0) {
                di = 0.5;
                x = x0 + di * (x1 - x0);
                if (x == 0.0)
                    goto under;
            }
            y = incbet(a, b, x);
            yp = (x1 - x0) / (x1 + x0);
            if (fabs(yp) < dithresh) {
                goto newt;
            }
            yp = (y - y0) / y0;
            if (fabs(yp) < dithresh) {
                goto newt;
            }
        }
        if (y < y0) {
            x0 = x;
            yl = y;
            if (dir < 0) {
                dir = 0;
                di = 0.5;
            } else if (dir > 3) {
                di = 1.0 - (1.0 - di) * (1.0 - di);
            } else if (dir > 1) {
                di = 0.5 * di + 0.5;
            } else {
                di = (y0 - y) / (yh - yl);
            }
            dir += 1;
            if (x0 > 0.75) {
                if (rflg == 1) {
                    rflg = 0;
                    a = aa;
                    b = bb;
                    y0 = yy0;
                } else {
                    rflg = 1;
                    a = bb;
                    b = aa;
                    y0 = 1.0 - yy0;
                }
                x = 1.0 - x;
                y = incbet(a, b, x);
                x0 = 0.0;
                yl = 0.0;
                x1 = 1.0;
                yh = 1.0;
                goto ihalve;
            }
        }
        else {
            x1 = x;
            if (rflg == 1 && x1 < MACHEP) {
                x = 0.0;
                goto done;
            }
            yh = y;
            if (dir > 0) {
                dir = 0;
                di = 0.5;
            } else if (dir < -3) {
                di = di * di;
            } else if (dir < -1) {
                di = 0.5 * di;
            } else {
                di = (y - y0) / (yh - yl);
            }
            dir -= 1;
        }
    }
    if (x0 >= 1.0) {
        x = 1.0 - MACHEP;
        goto done;
    }
    if (x <= 0.0) {
under:
        x = 0.0;
        goto done;
    }

newt:

    if (nflg) {
        goto done;
    }
    nflg = 1;
    lgm = lgam(a + b) - lgam(a) - lgam(b);

    for (i = 0; i < 8; i++) {
        /* Compute the function at this point. */
        if (i != 0) {
            y = incbet(a, b, x);
        }
        if (y < yl) {
            x = x0;
            y = yl;
        } else if (y > yh) {
            x = x1;
            y = yh;
        } else if (y < y0) {
            x0 = x;
            yl = y;
        } else {
            x1 = x;
            yh = y;
        }
        if (x == 1.0 || x == 0.0)
            break;
        /* Compute the derivative of the function at this point. */
        d = (a - 1.0) * log(x) + (b - 1.0) * log(1.0 - x) + lgm;
        if (d < MINLOG) {
            goto done;
        }
        if (d > MAXLOG) {
            break;
        }
        d = exp(d);
        /* Compute the step to the next approximation of x. */
        d = (y - y0) / d;
        xt = x - d;
        if (xt <= x0) {
            y = (x - x0) / (x1 - x0);
            xt = x0 + 0.5 * y * (x - x0);
            if (xt <= 0.0) {
                break;
            }
        }
        if (xt >= x1) {
            y = (x1 - x) / (x1 - x0);
            xt = x1 - 0.5 * y * (x1 - x);
            if (xt >= 1.0) {
                break;
            }
        }
        x = xt;
        if (fabs(d / x) < 128.0 * MACHEP) {
            goto done;
        }
    }
    /* Did not converge.  */
    dithresh = 256.0 * MACHEP;
    goto ihalve;

done:

    if (rflg) {
        if (x <= MACHEP)
            x = 1.0 - MACHEP;
        else
            x = 1.0 - x;
    }
    return x;
}

"""

incbi_preamble = incbet_preamble + incbi_definition


betaincinv = _core.create_ufunc(
    "cupyx_scipy_betaincinv",
    ("fff->f", "ddd->d"),
    "out0 = out0_type(incbi(in0, in1, in2));",
    preamble=incbi_preamble,
    doc="""Inverse of the incomplete beta function.

    Parameters
    ----------
    a, b : cupy.ndarray
        Positive, real-valued parameters
    y : cupy.ndarray
        Real-valued input.
    out : ndarray, optional
        Optional output array for the function result

    Returns
    -------
    scalar or ndarray
        Value of the inverse of the incomplete beta function

    See Also
    --------
    :func:`scipy.special.betaincinv`

    """,
)



from cupy import _core


logit_definition = """
template <typename T>
static __device__ T logit(T x) {
    x /= 1 - x;
    return log(x);
}

"""


logit = _core.create_ufunc(
    'cupy_logit',
    ('e->f', 'f->f', 'd->d'),
    'out0 = logit(in0)',
    preamble=logit_definition,
    doc='''Logit function.

    Args:
        x (cupy.ndarray): input data

    Returns:
        cupy.ndarray: values of logit(x)

    .. seealso:: :data:`scipy.special.logit`

    ''')


expit_definition = """
template <typename T>
static __device__ T expit(T x) {
    return 1 / (1 + exp(-x));
}

"""


expit = _core.create_ufunc(
    'cupy_expit',
    ('e->f', 'f->f', 'd->d'),
    'out0 = expit(in0)',
    preamble=expit_definition,
    doc='''Logistic sigmoid function (expit).

    Args:
        x (cupy.ndarray): input data (must be real)

    Returns:
        cupy.ndarray: values of expit(x)

    .. seealso:: :data:`scipy.special.expit`

    .. note::
        expit is the inverse of logit.

    ''')


# log_expit implemented based on log1p as in SciPy's scipy/special/_logit.h

log_expit_definition = """
template <typename T>
static __device__ T log_expit(T x)
{
    if (x < 0.0) {
        return x - log1p(exp(x));
    } else {
        return -log1p(exp(-x));
    }
}

"""


log_expit = _core.create_ufunc(
    'cupy_log_expit',
    ('e->f', 'f->f', 'd->d'),
    'out0 = log_expit(in0)',
    preamble=log_expit_definition,
    doc='''Logarithm of the logistic sigmoid function.

    Args:
        x (cupy.ndarray): input data (must be real)

    Returns:
        cupy.ndarray: values of log(expit(x))

    .. seealso:: :data:`scipy.special.log_expit`

    .. note::
        The function is mathematically equivalent to ``log(expit(x))``, but
        is formulated to avoid loss of precision for inputs with large
        (positive or negative) magnitude.
    ''')


boxcox_definition = """
static __device__ double boxcox(double x, double lmbda) {
    // if lmbda << 1 and log(x) < 1.0, the lmbda*log(x) product can lose
    // precision, furthermore, expm1(x) == x for x < eps.
    // For doubles, the range of log is -744.44 to +709.78, with eps being
    // the smallest value produced.  This range means that we will have
    // abs(lmbda)*log(x) < eps whenever abs(lmbda) <= eps/-log(min double)
    // which is ~2.98e-19.
    if (fabs(lmbda) < 1e-19) {
        return log(x);
    } else {
        return expm1(lmbda * log(x)) / lmbda;
    }
}

"""


boxcox = _core.create_ufunc(
    'cupy_boxcox',
    ('ee->f', 'ff->f', 'dd->d'),
    'out0 = out0_type(boxcox(in0, in1))',
    preamble=boxcox_definition,
    doc='''Compute the Box-Cox transformation.

    Args:
        x (cupy.ndarray): input data (must be real)

    Returns:
        cupy.ndarray: values of boxcox(x)

    .. seealso:: :data:`scipy.special.boxcox`

    ''')


boxcox1p_definition = """
static __device__ double boxcox1p(double x, double lmbda) {
    // The argument given above in boxcox applies here with the modification
    // that the smallest value produced by log1p is the minimum representable
    // value, rather than eps.  The second condition here prevents underflow
    // when log1p(x) is < eps.
    double lgx = log1p(x);
    if ((fabs(lmbda) < 1e-19)
        || ((fabs(lgx) < 1e-289) && (fabs(lmbda) < 1e273))) {
        return lgx;
    } else {
        return expm1(lmbda * lgx) / lmbda;
    }
}

"""


boxcox1p = _core.create_ufunc(
    'cupy_boxcox1p',
    ('ee->f', 'ff->f', 'dd->d'),
    'out0 = out0_type(boxcox1p(in0, in1))',
    preamble=boxcox1p_definition,
    doc='''Compute the Box-Cox transformation op 1 + `x`.

    Args:
        x (cupy.ndarray): input data (must be real)

    Returns:
        cupy.ndarray: values of boxcox1p(x)

    .. seealso:: :data:`scipy.special.boxcox1p`

    ''')


inv_boxcox_definition = """
static __device__ double inv_boxcox(double x, double lmbda) {
    if (lmbda == 0.0) {
        return exp(x);
    } else {
        return exp(log1p(lmbda * x) / lmbda);
    }
}

"""


inv_boxcox = _core.create_ufunc(
    'cupy_inv_boxcox',
    ('ee->f', 'ff->f', 'dd->d'),
    'out0 = out0_type(inv_boxcox(in0, in1))',
    preamble=inv_boxcox_definition,
    doc='''Compute the Box-Cox transformation.

    Args:
        x (cupy.ndarray): input data (must be real)

    Returns:
        cupy.ndarray: values of inv_boxcox(x)

    .. seealso:: :data:`scipy.special.inv_boxcox`

    ''')


inv_boxcox1p_definition = """
static __device__ double inv_boxcox1p(double x, double lmbda) {
    if (lmbda == 0.0) {
        return expm1(x);
    } else if (fabs(lmbda * x) < 1e-154) {
        return x;
    } else {
        return expm1(log1p(lmbda * x) / lmbda);
    }
}

"""


inv_boxcox1p = _core.create_ufunc(
    'cupy_inv_boxcox1p',
    ('ee->f', 'ff->f', 'dd->d'),
    'out0 = out0_type(inv_boxcox1p(in0, in1))',
    preamble=inv_boxcox1p_definition,
    doc='''Compute the Box-Cox transformation op 1 + `x`.

    Args:
        x (cupy.ndarray): input data (must be real)

    Returns:
        cupy.ndarray: values of inv_boxcox1p(x)

    .. seealso:: :data:`scipy.special.inv_boxcox1p`
''')



# Copyright (c) 2019-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE


import cupy


_real_cepstrum_kernel = cupy.ElementwiseKernel(
    "T spectrum",
    "T output",
    """
    output = log( abs( spectrum ) );
    """,
    "_real_cepstrum_kernel",
    options=("-std=c++11",),
)


def real_cepstrum(x, n=None, axis=-1):
    r"""
    Calculates the real cepstrum of an input sequence x where the cepstrum is
    defined as the inverse Fourier transform of the log magnitude DFT
    (spectrum) of a signal. It's primarily used for source/speaker separation
    in speech signal processing

    Parameters
    ----------
    x : ndarray
        Input sequence, if x is a matrix, return cepstrum in direction of axis
    n : int
        Size of Fourier Transform; If none, will use length of input array
    axis: int
        Direction for cepstrum calculation

    Returns
    -------
    ceps : ndarray
        Complex cepstrum result
    """
    x = cupy.asarray(x)
    spectrum = cupy.fft.fft(x, n=n, axis=axis)
    spectrum = _real_cepstrum_kernel(spectrum)
    return cupy.fft.ifft(spectrum, n=n, axis=axis).real


_complex_cepstrum_kernel = cupy.ElementwiseKernel(
    "C spectrum, raw T unwrapped",
    "C output, T ndelay",
    """
    ndelay = round( unwrapped[center] / M_PI );
    const T temp { unwrapped[i] - ( M_PI * ndelay * i / center ) };

    output = log( abs( spectrum ) ) + C( 0, temp );
    """,
    "_complex_cepstrum_kernel",
    options=("-std=c++11",),
    return_tuple=True,
    loop_prep="const int center { static_cast<int>( 0.5 * \
        ( _ind.size() + 1 ) ) };",
)


def complex_cepstrum(x, n=None, axis=-1):
    r"""
    Calculates the complex cepstrum of a real valued input sequence x
    where the cepstrum is defined as the inverse Fourier transform
    of the log magnitude DFT (spectrum) of a signal. It's primarily
    used for source/speaker separation in speech signal processing.

    The input is altered to have zero-phase at pi radians (180 degrees)
    Parameters
    ----------
    x : ndarray
        Input sequence, if x is a matrix, return cepstrum in direction of axis
    n : int
       Size of Fourier Transform; If none, will use length of input array
    axis: int
        Direction for cepstrum calculation
    Returns
    -------
    ceps : ndarray
        Complex cepstrum result
    """
    x = cupy.asarray(x)
    spectrum = cupy.fft.fft(x, n=n, axis=axis)
    unwrapped = cupy.unwrap(cupy.angle(spectrum))
    log_spectrum, ndelay = _complex_cepstrum_kernel(spectrum, unwrapped)
    ceps = cupy.fft.ifft(log_spectrum, n=n, axis=axis).real

    return ceps, ndelay


_inverse_complex_cepstrum_kernel = cupy.ElementwiseKernel(
    "C log_spectrum, int32 ndelay, float64 pi",
    "C spectrum",
    """
    const double wrapped { log_spectrum.imag() + M_PI * ndelay * i / center };

    spectrum = exp( C( log_spectrum.real(), wrapped ) )
    """,
    "_inverse_complex_cepstrum_kernel",
    options=("-std=c++11",),
    loop_prep="const double center { 0.5 * ( _ind.size() + 1 ) };",
)


def inverse_complex_cepstrum(ceps, ndelay):
    r"""Compute the inverse complex cepstrum of a real sequence.
    ceps : ndarray
        Real sequence to compute inverse complex cepstrum of.
    ndelay: int
        The amount of samples of circular delay added to `x`.
    Returns
    -------
    x : ndarray
        The inverse complex cepstrum of the real sequence `ceps`.
    The inverse complex cepstrum is given by
    .. math:: x[n] = F^{-1}\left{\exp(F(c[n]))\right}
    where :math:`c_[n]` is the input signal and :math:`F` and :math:`F_{-1}
    are respectively the forward and backward Fourier transform.
    """
    ceps = cupy.asarray(ceps)
    log_spectrum = cupy.fft.fft(ceps)
    spectrum = _inverse_complex_cepstrum_kernel(log_spectrum, ndelay, cupy.pi)
    iceps = cupy.fft.ifft(spectrum).real

    return iceps


_minimum_phase_kernel = cupy.ElementwiseKernel(
    "T ceps",
    "T window",
    """
    if ( !i ) {
        window = ceps;
    } else if ( i < bend ) {
        window = ceps * 2.0;
    } else if ( i == bend ) {
        window = ceps * ( 1 - odd );
    } else {
        window = 0;
    }
    """,
    "_minimum_phase_kernel",
    options=("-std=c++11",),
    loop_prep="const bool odd { _ind.size() & 1 }; \
               const int bend { static_cast<int>( 0.5 * \
                    ( _ind.size() + odd ) ) };",
)


def minimum_phase(x, n=None):
    r"""Compute the minimum phase reconstruction of a real sequence.
    x : ndarray
        Real sequence to compute the minimum phase reconstruction of.
    n : {None, int}, optional
        Length of the Fourier transform.
    Compute the minimum phase reconstruction of a real sequence using the
    real cepstrum.
    Returns
    -------
    m : ndarray
        The minimum phase reconstruction of the real sequence `x`.
    """
    if n is None:
        n = len(x)
    ceps = real_cepstrum(x, n=n)
    window = _minimum_phase_kernel(ceps)
    m = cupy.fft.ifft(cupy.exp(cupy.fft.fft(window))).real

    return m



import collections
import os
import subprocess
import sys

import pytest


# Enable `testdir` fixture to test `cupy.testing`.
# `pytest_plugins` cannot be locally configured. See also
# https://docs.pytest.org/en/stable/deprecations.html#pytest-plugins-in-non-top-level-conftest-files
pytest_plugins = ['pytester']


def _is_pip_installed():
    try:
        import pip  # NOQA
        return True
    except ImportError:
        return False


def _is_in_ci():
    ci_name = os.environ.get('CUPY_CI', '')
    return ci_name != ''


def pytest_configure(config):
    # Print installed packages
    if _is_in_ci() and _is_pip_installed():
        print("***** Installed packages *****", flush=True)
        subprocess.check_call([sys.executable, '-m', 'pip', 'freeze', '--all'])
    if config.pluginmanager.hasplugin("xdist"):
        config.pluginmanager.register(DeferPlugin())


# https://docs.pytest.org/en/latest/how-to/writing_hook_functions.html#optionally-using-hooks-from-3rd-party-plugins
class DeferPlugin:
    """Simple plugin to defer pytest-xdist hook functions."""

    # Edit the environment variable `CUDA_VISIBLE_DEVICES` for each session.
    # Cannot use `pytest_configure_node` nor `pytest_testnodeready` hook,
    # because they are called in the `master` node (process).
    # See also https://github.com/pytest-dev/pytest-xdist/issues/179.
    @pytest.fixture(autouse=True, scope='session')
    def _rotate_cuda_visible_devices(self, worker_id):
        if worker_id == 'master':
            # `worker_id` can be `master` if `pytest-xdist` is installed and
            # run without `-n` option.
            return

        n_gpu = os.environ.get('CUPY_TEST_GPU_LIMIT')
        if n_gpu is None:
            print('Tip: when using pytest-xdist, you can automatically rotate'
                  ' CUDA_VISIBLE_DEVICES for each test worker by setting'
                  ' CUPY_TEST_GPU_LIMIT environment variable.')
            return
        n_gpu = int(n_gpu)

        assert worker_id.startswith('gw')
        w = int(worker_id[2:])

        devices = os.environ.get('CUDA_VISIBLE_DEVICES')
        if devices is None:
            devices = [str(k) for k in range(n_gpu)]
        else:
            devices = devices.split(',')[:n_gpu]
        devices = collections.deque(devices)
        devices.rotate(w)
        devices = ','.join(devices)
        os.environ['CUDA_VISIBLE_DEVICES'] = devices
        # With PyTest's default, the print will be shown as
        # "--- Captured stdout setup ---" on failure.
        print(f'CUDA_VISIBLE_DEVICES={devices}')


if int(os.environ.get('CUPY_ENABLE_UMP', 0)) != 0:
    # Make sure malloc is used in a stream-ordered fashion
    import cupy as cp
    cp.cuda.set_allocator(cp.cuda.MemoryPool(
        cp.cuda.memory.malloc_system).malloc)

    import cupy._core.numpy_allocator as ac
    import numpy_allocator
    import ctypes
    lib = ctypes.CDLL(ac.__file__)

    class my_allocator(metaclass=numpy_allocator.type):
        _calloc_ = ctypes.addressof(lib._calloc)
        _malloc_ = ctypes.addressof(lib._malloc)
        _realloc_ = ctypes.addressof(lib._realloc)
        _free_ = ctypes.addressof(lib._free)
    my_allocator.__enter__()



"""
Implements the binom function from scipy.

This is essentially a CUDA C++ adaptation of existing scipy code, available at:
https://github.com/scipy/scipy/blob/v1.10.1/scipy/special/orthogonal_eval.pxd
"""

from cupy import _core
from cupyx.scipy.special._beta import (
    beta_preamble,
    lbeta_symp_definition,
    lgam_sgn_definition,
    beta_definition,
    lbeta_definition
)
from cupyx.scipy.special._digamma import polevl_definition
from cupyx.scipy.special._gamma import gamma_definition
from cupyx.scipy.special._gammainc import p1evl_definition

binom_definition = """
__device__ double binom(double n, double k) {

    // Check if n is negative integer
    if (n < 0 && n == trunc(n)) {
        return CUDART_NAN;
    }
    // Check if k is infinite
    if (k == CUDART_INF) {
        return CUDART_NAN;
    }

    // k is integer and n is not very small
    if (k == floor(k) && (fabs(n) > 1.0e-8 || n == 0)) {
        // Reduce k by symmetry if n also integer
        if (n == floor(n) && k > n/2 && n > 0) {
            k = n - k;
        }
        // Manually multiply if k between 0, 20
        if (0 <= k && k < 20) {
            double num = 1.0;
            double den = 1.0;
            for (int i = 1; i < k+.5; i++) {
                num *= i + n - k;
                den *= i;
                // Resize num, den if num becomes too big
                if (fabs(num) > 1.0e50) {
                    num /= den;
                    den = 1.0;
                }
            }
            return num / den;
        }
    }

    if (n >= k * 1e10 && k > 0) {
        // Prevent overflow
        return exp(-lbeta(1 + n - k, 1 + k) - log(n + 1));
    } else if (k > 1e8 * fabs(n)) {
        double num = Gamma(1 + n) / fabs(k) + Gamma(1 + n) * n / (2 * k*k);
        num /= M_PI * pow(fabs(k), n);
        double kfloor = floor(k);
        if (k > 0) {
            int sgn;
            double dk;
            if (int(kfloor) == kfloor) {
                dk = k - kfloor;
                sgn = int(kfloor) % 2 == 0 ? 1 : -1;
            } else {
                dk = k;
                sgn = 1;
            }
            return num * sin((dk-n)*M_PI) * sgn;
        } else {
            if (int(kfloor) == kfloor) {
                return 0;
            } else {
                return num * sin(k*M_PI);
            }
        }
    } else {
        return 1/(n + 1)/beta(1 + n - k, 1 + k);
    }
}
"""

binom = _core.create_ufunc(
    "cupyx_scipy_binom",
    ("ff->f", "dd->d"),
    "out0 = out0_type(binom(in0, in1));",
    preamble=(
        gamma_definition +
        beta_preamble +
        p1evl_definition +
        polevl_definition +
        lgam_sgn_definition +
        lbeta_symp_definition +
        lbeta_definition +
        beta_definition +
        binom_definition
    ),
    doc="""Binomial coefficient

    Parameters
    ----------
    n, k : cupy.ndarray
        Real-valued parameters to nCk
    out : ndarray, optional
        Optional output array for the function result

    Returns
    -------
    scalar or ndarray
        Value of binomial coefficient

    .. seealso:: :func:`scipy.special.binom`

    """,
)



"""basic special functions


cotdg and tandg implementations are adapted from the following SciPy code:

https://github.com/scipy/scipy/blob/master/scipy/special/cephes/tandg.c

radian is from

https://github.com/scipy/scipy/blob/master/scipy/special/cephes/sindg.c

cosm1 is from

https://github.com/scipy/scipy/blob/main/scipy/special/cephes/unity.c

polevl is from

https://github.com/scipy/scipy/blob/main/scipy/special/cephes/polevl.h


Cephes Math Library Release 2.0:  April, 1987
Copyright 1984, 1987 by Stephen L. Moshier
Direct inquiries to 30 Frost Street, Cambridge, MA 02140
"""

from cupy import _core


# Note: cast complex<single> to complex<double> or tests fail tolerance
log1p = _core.create_ufunc(
    'cupyx_scipy_log1p',
    (('f->f', 'out0 = log1pf(in0)'),
     ('F->F', 'out0 = out0_type(log1p((complex<double>)in0))'),
     'd->d', 'D->D'),
    'out0 = log1p(in0);',
    doc="""Elementwise function for scipy.special.log1p

    Calculates log(1 + x) for use when `x` is near zero.

    Notes
    -----
    This implementation currently does not support complex-valued `x`.

    .. seealso:: :meth:`scipy.special.log1p`

    """,
)


cbrt = _core.create_ufunc(
    'cupyx_scipy_special_cbrt',
    (('f->f', 'out0 = cbrtf(in0)'), 'd->d'),
    'out0 = cbrt(in0)',
    doc='''Cube root.

    .. seealso:: :meth:`scipy.special.cbrt`

    ''')


exp2 = _core.create_ufunc(
    'cupyx_scipy_special_exp2',
    (('f->f', 'out0 = exp2f(in0)'), 'd->d'),
    'out0 = exp2(in0)',
    doc='''Computes ``2**x``.

    .. seealso:: :meth:`scipy.special.exp2`

    ''')


exp10 = _core.create_ufunc(
    'cupyx_scipy_special_exp10',
    (('f->f', 'out0 = exp10f(in0)'), 'd->d'),
    'out0 = exp10(in0)',
    doc='''Computes ``10**x``.

    .. seealso:: :meth:`scipy.special.exp10`

    ''')


expm1 = _core.create_ufunc(
    'cupyx_scipy_special_expm1',
    (('f->f', 'out0 = expm1f(in0)'), 'd->d', 'F->F', 'D->D'),
    'out0 = expm1(in0)',
    doc='''Computes ``exp(x) - 1``.

    .. seealso:: :meth:`scipy.special.expm1`

    ''')

exprel = _core.create_ufunc(
    'cupyx_scipy_special_exprel',
    (('f->f'), 'd->d', 'F->F', 'D->D'),
    'out0 = abs(in0) >= 1e-16 ? expm1(in0) / in0 : 1',
    doc='''Computes ``(exp(x) - 1) / x``.

    .. seealso:: :meth:`scipy.special.exprel`

    ''')

cosm1_implementation = """
//Define from npy_math.h
//https://github.com/numpy/numpy/blob/main/numpy/core/include/numpy/npy_math.h
#define NPY_PI_4      0.785398163397448309615660845819875721  /* pi/4 */
__constant__ double coscof[] = {
    4.7377507964246204691685E-14,
    -1.1470284843425359765671E-11,
    2.0876754287081521758361E-9,
    -2.7557319214999787979814E-7,
    2.4801587301570552304991E-5,
    -1.3888888888888872993737E-3,
    4.1666666666666666609054E-2,
};

__device__ static double polevl(double x, const double coef[], int N)
{
    double ans;
    int i;
    const double *p;

    p = coef;
    ans = *p++;
    i = N;

    do
        ans = ans * x + *p++;
    while (--i);

    return (ans);
}

__device__ static double cosm1(double x)
{
    double xx;

    if ((x < -NPY_PI_4) || (x > NPY_PI_4))
        return (cos(x) - 1.0);
    xx = x * x;
    xx = -0.5 * xx + xx * xx * polevl(xx, coscof, 6);
    return xx;
}
"""

cosm1 = _core.create_ufunc(
    'cupyx_scipy_special_cosm1', ('f->f', 'd->d'),
    'out0 = cosm1(in0)',
    preamble=cosm1_implementation,
    doc='''Computes ``cos(x) - 1``.

    .. seealso:: :meth:`scipy.special.cosm1`

    ''')

pi180_preamble = """
    __constant__ double PI180 = 1.74532925199432957692E-2;  // pi/180
"""

cosdg = _core.create_ufunc(
    'cupyx_scipy_special_cosdg',
    (('f->f', 'out0 = cosf(PI180 * in0)'), 'd->d'),
    'out0 = cos(PI180 * in0)',
    preamble=pi180_preamble,
    doc='''Cosine of x with x in degrees.

    .. seealso:: :meth:`scipy.special.cosdg`

    ''')


sindg = _core.create_ufunc(
    'cupyx_scipy_special_sindg',
    (('f->f', 'out0 = sinf(PI180 * in0)'), 'd->d'),
    'out0 = sin(PI180 * in0)',
    preamble=pi180_preamble,
    doc='''Sine of x with x in degrees.

    .. seealso:: :meth:`scipy.special.sindg`

    ''')


tancot_implementation = pi180_preamble + """


// include for CUDART_INF
#include <cupy/math_constants.h>

__constant__ double  lossth = 1.0e14;

__device__ static double tancot(double, int);

__device__ static double tandg(double x)
{
    return tancot(x, 0);
}


__device__ static double cotdg(double x)
{
    return tancot(x, 1);
}


__device__ static double tancot(double xx, int cotflg)
{
    double x;
    int sign;

    /* make argument positive but save the sign */
    if (xx < 0) {
        x = -xx;
        sign = -1;
    }
    else {
        x = xx;
        sign = 1;
    }

    if (x > lossth) {
        // sf_error("tandg", SF_ERROR_NO_RESULT, NULL);
        return 0.0;
    }

    /* modulo 180 */
    x = x - 180.0 * floor(x / 180.0);
    if (cotflg) {
        if (x <= 90.0) {
            x = 90.0 - x;
        } else {
            x = x - 90.0;
            sign *= -1;
        }
    } else {
        if (x > 90.0) {
            x = 180.0 - x;
            sign *= -1;
        }
    }
    if (x == 0.0) {
        return 0.0;
    }
    else if (x == 45.0) {
        return sign * 1.0;
    }
    else if (x == 90.0) {
        // sf_error((cotflg ? "cotdg" : "tandg"), SF_ERROR_SINGULAR, NULL);
        return CUDART_INF;
    }
    /* x is now transformed into [0, 90) */
    return sign * tan(x * PI180);
}
"""

tandg = _core.create_ufunc(
    'cupyx_scipy_special_tandg', ('f->f', 'd->d'),
    'out0 = tandg(in0)',
    preamble=tancot_implementation,
    doc='''Tangent of x with x in degrees.

    .. seealso:: :meth:`scipy.special.tandg`

    ''')


cotdg = _core.create_ufunc(
    'cupyx_scipy_special_cotdg', ('f->f', 'd->d'),
    'out0 = cotdg(in0)',
    preamble=tancot_implementation,
    doc='''Cotangent of x with x in degrees.

    .. seealso:: :meth:`scipy.special.cotdg`

    ''')

radian_implementation = """
/* 1 arc second, in radians*/
__constant__ double P64800 =
    4.848136811095359935899141023579479759563533023727e-6;

template <typename T>
__device__ T radian(T d, T m, T s)
{
    return (((d * 60.0 + m) * 60.0 + s) * P64800);
}
"""

radian = _core.create_ufunc(
    'cupyx_scipy_special_radian', ('fff->f', 'ddd->d'),
    'out0 = radian(in0, in1, in2)',
    preamble=radian_implementation,
    doc='''Degrees, minutes, seconds to radians:

    .. seealso:: :meth:`scipy.special.radian`

    ''')



from pathlib import Path

import pytest


@pytest.mark.parametrize("t", map(str, Path(__file__).parent.glob("**/*.pyi")))
def test_run(t: Path) -> None:
    assert not t.endswith("._numpy.pyi")

    with open(t) as f:
        lines = f.readlines()

    exec_locals = {}
    for _lineno, _line in enumerate(lines, start=1):
        if "# E: " not in _line:
            try:
                exec(_line, None, exec_locals)
            except Exception:
                print(f"{t}:{_lineno} {_line}")
                raise



import unittest

import numpy

import cupy
from cupy import testing
from cupy.lib import stride_tricks

import pytest


class TestAsStrided(unittest.TestCase):
    def test_as_strided(self):
        a = cupy.array([1, 2, 3, 4])
        a_view = stride_tricks.as_strided(
            a, shape=(2,), strides=(2 * a.itemsize,))
        expected = cupy.array([1, 3])
        testing.assert_array_equal(a_view, expected)

        a = cupy.array([1, 2, 3, 4])
        a_view = stride_tricks.as_strided(
            a, shape=(3, 4), strides=(0, 1 * a.itemsize))
        expected = cupy.array([[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]])
        testing.assert_array_equal(a_view, expected)

    @testing.numpy_cupy_array_equal()
    def test_rolling_window(self, xp):
        a = testing.shaped_arange((3, 4), xp)
        a_rolling = rolling_window(a, 2, 0)
        return a_rolling


class TestSlidingWindowView(unittest.TestCase):
    @testing.numpy_cupy_array_equal()
    def test_1d(self, xp):
        arr = testing.shaped_arange((3, 4), xp)
        window_size = 2
        arr_view = xp.lib.stride_tricks.sliding_window_view(
            arr, window_size, 0)
        assert arr_view.strides == (16, 4, 16)
        return arr_view

    @testing.numpy_cupy_array_equal()
    def test_2d(self, xp):
        arr = testing.shaped_arange((3, 4), xp)
        window_shape = (2, 2)
        arr_view = xp.lib.stride_tricks.sliding_window_view(
            arr, window_shape=window_shape
        )
        assert arr_view.strides == (16, 4, 16, 4)
        return arr_view

    @testing.numpy_cupy_array_equal()
    def test_2d_with_axis(self, xp):
        arr = testing.shaped_arange((3, 4), xp)
        window_shape = 3
        axis = 1
        arr_view = xp.lib.stride_tricks.sliding_window_view(
            arr, window_shape, axis)
        assert arr_view.strides == (16, 4, 4)
        return arr_view

    @testing.numpy_cupy_array_equal()
    def test_2d_multi_axis(self, xp):
        arr = testing.shaped_arange((3, 4), xp)
        window_shape = (2, 3)
        axis = (0, 1)
        arr_view = xp.lib.stride_tricks.sliding_window_view(
            arr, window_shape, axis)
        assert arr_view.strides == (16, 4, 16, 4)
        return arr_view

    def test_0d(self):
        for xp in (numpy, cupy):
            # Create a 0-D array (scalar) for testing
            arr = xp.array(42)
            # Sliding window with window size 1
            window_size = 1

            # Test if the correct ValueError is raised!
            with pytest.raises(ValueError, match="axis 0 is out of bounds"):
                xp.lib.stride_tricks.sliding_window_view(arr, window_size, 0)

    def test_window_shape_axis_length_mismatch(self):
        for xp in (numpy, cupy):
            x = xp.arange(24).reshape((2, 3, 4))
            window_shape = (2, 2)
            axis = None

            # Test if ValueError is raised when len(window_shape) != len(axis)
            with pytest.raises(ValueError, match="Since axis is `None`"):
                xp.lib.stride_tricks.sliding_window_view(x, window_shape, axis)

    @testing.numpy_cupy_array_equal()
    def test_arraylike_input(self, xp):
        x = [0., 1., 2., 3., 4.]
        arr_view = xp.lib.stride_tricks.sliding_window_view(x, 2)
        assert arr_view.strides == (8, 8)
        return arr_view

    def test_writeable_views_not_supported(self):
        x = cupy.arange(24).reshape((2, 3, 4))
        window_shape = (2, 2)
        axis = None
        writeable = True

        with self.assertRaises(NotImplementedError):
            stride_tricks.sliding_window_view(
                x, window_shape, axis, writeable=writeable
            )


def rolling_window(a, window, axis=-1):
    """
    Make an ndarray with a rolling window along axis.
    This function is taken from https://github.com/numpy/numpy/pull/31
    but slightly modified to accept axis option.
    """
    a = numpy.swapaxes(a, axis, -1)
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    if isinstance(a, numpy.ndarray):
        rolling = numpy.lib.stride_tricks.as_strided(
            a, shape=shape, strides=strides)
    elif isinstance(a, cupy.ndarray):
        rolling = stride_tricks.as_strided(a, shape=shape, strides=strides)
    return rolling.swapaxes(-2, axis)



"""
Routines for manipulating partial fraction expansions.
"""

import cupy


def roots(arr):
    """np.roots replacement. XXX: calls into NumPy, then converts back.
    """
    import numpy as np

    arr = cupy.asarray(arr).get()
    return cupy.asarray(np.roots(arr))


def poly(A):
    """np.poly replacement for 2D A. Otherwise, use cupy.poly."""
    sh = A.shape
    if not (len(sh) == 2 and sh[0] == sh[1] and sh[0] != 0):
        raise ValueError("input must be a non-empty square 2d array.")

    import numpy as np

    seq_of_zeros = np.linalg.eigvals(A.get())

    dt = seq_of_zeros.dtype
    a = np.ones((1,), dtype=dt)
    for zero in seq_of_zeros:
        a = np.convolve(a, np.r_[1, -zero], mode='full')

    if issubclass(a.dtype.type, cupy.complexfloating):
        # if complex roots are all complex conjugates, the roots are real.
        roots = np.asarray(seq_of_zeros, dtype=complex)
        if np.all(np.sort(roots) == np.sort(roots.conjugate())):
            a = a.real.copy()

    return cupy.asarray(a)


def _cmplx_sort(p):
    """Sort roots based on magnitude.
    """
    indx = cupy.argsort(cupy.abs(p))
    return cupy.take(p, indx, 0), indx


# np.polydiv clone
def _polydiv(u, v):
    u = cupy.atleast_1d(u) + 0.0
    v = cupy.atleast_1d(v) + 0.0
    # w has the common type
    w = u[0] + v[0]
    m = len(u) - 1
    n = len(v) - 1
    scale = 1. / v[0]
    q = cupy.zeros((max(m - n + 1, 1),), w.dtype)
    r = u.astype(w.dtype)
    for k in range(0, m-n+1):
        d = scale * r[k]
        q[k] = d
        r[k:k + n + 1] -= d * v
    while cupy.allclose(r[0], 0, rtol=1e-14) and (r.shape[-1] > 1):
        r = r[1:]
    return q, r


def unique_roots(p, tol=1e-3, rtype='min'):
    """Determine unique roots and their multiplicities from a list of roots.

    Parameters
    ----------
    p : array_like
        The list of roots.
    tol : float, optional
        The tolerance for two roots to be considered equal in terms of
        the distance between them. Default is 1e-3. Refer to Notes about
        the details on roots grouping.
    rtype : {'max', 'maximum', 'min', 'minimum', 'avg', 'mean'}, optional
        How to determine the returned root if multiple roots are within
        `tol` of each other.

          - 'max', 'maximum': pick the maximum of those roots
          - 'min', 'minimum': pick the minimum of those roots
          - 'avg', 'mean': take the average of those roots

        When finding minimum or maximum among complex roots they are compared
        first by the real part and then by the imaginary part.

    Returns
    -------
    unique : ndarray
        The list of unique roots.
    multiplicity : ndarray
        The multiplicity of each root.

    See Also
    --------
    scipy.signal.unique_roots

    Notes
    -----
    If we have 3 roots ``a``, ``b`` and ``c``, such that ``a`` is close to
    ``b`` and ``b`` is close to ``c`` (distance is less than `tol`), then it
    doesn't necessarily mean that ``a`` is close to ``c``. It means that roots
    grouping is not unique. In this function we use "greedy" grouping going
    through the roots in the order they are given in the input `p`.

    This utility function is not specific to roots but can be used for any
    sequence of values for which uniqueness and multiplicity has to be
    determined. For a more general routine, see `numpy.unique`.

    """
    if rtype in ['max', 'maximum']:
        reduce = cupy.max
    elif rtype in ['min', 'minimum']:
        reduce = cupy.min
    elif rtype in ['avg', 'mean']:
        reduce = cupy.mean
    else:
        raise ValueError("`rtype` must be one of "
                         "{'max', 'maximum', 'min', 'minimum', 'avg', 'mean'}")

    points = cupy.empty((p.shape[0], 2))
    points[:, 0] = cupy.real(p)
    points[:, 1] = cupy.imag(p)

    # Replacement for dist = cdist(points, points) to avoid needing `pylibraft`
    dist = cupy.linalg.norm(points[:, None, :] - points[None, :, :], axis=-1)

    p_unique = []
    p_multiplicity = []
    used = cupy.zeros(p.shape[0], dtype=bool)

    for i, ds in enumerate(dist):
        if used[i]:
            continue

        mask = (ds < tol) & ~used
        group = ds[mask]
        if group.size > 0:
            # print(j, ' : ', group, p[mask])
            p_unique.append(reduce(p[mask]))
            p_multiplicity.append(group.shape[0])
        used[mask] = True

    return cupy.asarray(p_unique), cupy.asarray(p_multiplicity)


def _compute_factors(roots, multiplicity, include_powers=False):
    """Compute the total polynomial divided by factors for each root."""
    current = cupy.array([1])
    suffixes = [current]
    for pole, mult in zip(roots[-1:0:-1], multiplicity[-1:0:-1]):
        monomial = cupy.r_[1, -pole]
        for _ in range(int(mult)):
            current = cupy.polymul(current, monomial)
        suffixes.append(current)
    suffixes = suffixes[::-1]

    factors = []
    current = cupy.array([1])
    for pole, mult, suffix in zip(roots, multiplicity, suffixes):
        monomial = cupy.r_[1, -pole]
        block = []
        for i in range(int(mult)):
            if i == 0 or include_powers:
                block.append(cupy.polymul(current, suffix))
            current = cupy.polymul(current, monomial)
        factors.extend(reversed(block))

    return factors, current


def _compute_residues(poles, multiplicity, numerator):
    denominator_factors, _ = _compute_factors(poles, multiplicity)
    numerator = numerator.astype(poles.dtype)

    residues = []
    for pole, mult, factor in zip(poles, multiplicity,
                                  denominator_factors):
        if mult == 1:
            residues.append(cupy.polyval(numerator, pole) /
                            cupy.polyval(factor, pole))
        else:
            numer = numerator.copy()
            monomial = cupy.r_[1, -pole]
            factor, d = _polydiv(factor, monomial)

            block = []
            for _ in range(int(mult)):
                numer, n = _polydiv(numer, monomial)
                r = n[0] / d[0]
                numer = cupy.polysub(numer, r * factor)
                block.append(r)

            residues.extend(reversed(block))

    return cupy.asarray(residues)


def invres(r, p, k, tol=1e-3, rtype='avg'):
    """Compute b(s) and a(s) from partial fraction expansion.

    If `M` is the degree of numerator `b` and `N` the degree of denominator
    `a`::

              b(s)     b[0] s**(M) + b[1] s**(M-1) + ... + b[M]
      H(s) = ------ = ------------------------------------------
              a(s)     a[0] s**(N) + a[1] s**(N-1) + ... + a[N]

    then the partial-fraction expansion H(s) is defined as::

               r[0]       r[1]             r[-1]
           = -------- + -------- + ... + --------- + k(s)
             (s-p[0])   (s-p[1])         (s-p[-1])

    If there are any repeated roots (closer together than `tol`), then H(s)
    has terms like::

          r[i]      r[i+1]              r[i+n-1]
        -------- + ----------- + ... + -----------
        (s-p[i])  (s-p[i])**2          (s-p[i])**n

    This function is used for polynomials in positive powers of s or z,
    such as analog filters or digital filters in controls engineering.  For
    negative powers of z (typical for digital filters in DSP), use `invresz`.

    Parameters
    ----------
    r : array_like
        Residues corresponding to the poles. For repeated poles, the residues
        must be ordered to correspond to ascending by power fractions.
    p : array_like
        Poles. Equal poles must be adjacent.
    k : array_like
        Coefficients of the direct polynomial term.
    tol : float, optional
        The tolerance for two roots to be considered equal in terms of
        the distance between them. Default is 1e-3. See `unique_roots`
        for further details.
    rtype : {'avg', 'min', 'max'}, optional
        Method for computing a root to represent a group of identical roots.
        Default is 'avg'. See `unique_roots` for further details.

    Returns
    -------
    b : ndarray
        Numerator polynomial coefficients.
    a : ndarray
        Denominator polynomial coefficients.

    See Also
    --------
    scipy.signal.invres
    residue, invresz, unique_roots

    """
    r = cupy.atleast_1d(r)
    p = cupy.atleast_1d(p)
    k = cupy.trim_zeros(cupy.atleast_1d(k), 'f')

    unique_poles, multiplicity = unique_roots(p, tol, rtype)
    factors, denominator = _compute_factors(unique_poles, multiplicity,
                                            include_powers=True)

    if len(k) == 0:
        numerator = 0
    else:
        numerator = cupy.polymul(k, denominator)

    for residue, factor in zip(r, factors):
        numerator = cupy.polyadd(numerator, residue * factor)

    return numerator, denominator


def invresz(r, p, k, tol=1e-3, rtype='avg'):
    """Compute b(z) and a(z) from partial fraction expansion.

    If `M` is the degree of numerator `b` and `N` the degree of denominator
    `a`::

                b(z)     b[0] + b[1] z**(-1) + ... + b[M] z**(-M)
        H(z) = ------ = ------------------------------------------
                a(z)     a[0] + a[1] z**(-1) + ... + a[N] z**(-N)

    then the partial-fraction expansion H(z) is defined as::

                 r[0]                   r[-1]
         = --------------- + ... + ---------------- + k[0] + k[1]z**(-1) ...
           (1-p[0]z**(-1))         (1-p[-1]z**(-1))

    If there are any repeated roots (closer than `tol`), then the partial
    fraction expansion has terms like::

             r[i]              r[i+1]                    r[i+n-1]
        -------------- + ------------------ + ... + ------------------
        (1-p[i]z**(-1))  (1-p[i]z**(-1))**2         (1-p[i]z**(-1))**n

    This function is used for polynomials in negative powers of z,
    such as digital filters in DSP.  For positive powers, use `invres`.

    Parameters
    ----------
    r : array_like
        Residues corresponding to the poles. For repeated poles, the residues
        must be ordered to correspond to ascending by power fractions.
    p : array_like
        Poles. Equal poles must be adjacent.
    k : array_like
        Coefficients of the direct polynomial term.
    tol : float, optional
        The tolerance for two roots to be considered equal in terms of
        the distance between them. Default is 1e-3. See `unique_roots`
        for further details.
    rtype : {'avg', 'min', 'max'}, optional
        Method for computing a root to represent a group of identical roots.
        Default is 'avg'. See `unique_roots` for further details.

    Returns
    -------
    b : ndarray
        Numerator polynomial coefficients.
    a : ndarray
        Denominator polynomial coefficients.

    See Also
    --------
    scipy.signal.invresz
    residuez, unique_roots, invres

    """
    r = cupy.atleast_1d(r)
    p = cupy.atleast_1d(p)
    k = cupy.trim_zeros(cupy.atleast_1d(k), 'b')

    unique_poles, multiplicity = unique_roots(p, tol, rtype)
    factors, denominator = _compute_factors(unique_poles, multiplicity,
                                            include_powers=True)

    if len(k) == 0:
        numerator = 0
    else:
        numerator = cupy.polymul(k[::-1], denominator[::-1])

    for residue, factor in zip(r, factors):
        numerator = cupy.polyadd(numerator, residue * factor[::-1])

    return numerator[::-1], denominator


def residue(b, a, tol=1e-3, rtype='avg'):
    """Compute partial-fraction expansion of b(s) / a(s).

    If `M` is the degree of numerator `b` and `N` the degree of denominator
    `a`::

              b(s)     b[0] s**(M) + b[1] s**(M-1) + ... + b[M]
      H(s) = ------ = ------------------------------------------
              a(s)     a[0] s**(N) + a[1] s**(N-1) + ... + a[N]

    then the partial-fraction expansion H(s) is defined as::

               r[0]       r[1]             r[-1]
           = -------- + -------- + ... + --------- + k(s)
             (s-p[0])   (s-p[1])         (s-p[-1])

    If there are any repeated roots (closer together than `tol`), then H(s)
    has terms like::

          r[i]      r[i+1]              r[i+n-1]
        -------- + ----------- + ... + -----------
        (s-p[i])  (s-p[i])**2          (s-p[i])**n

    This function is used for polynomials in positive powers of s or z,
    such as analog filters or digital filters in controls engineering.  For
    negative powers of z (typical for digital filters in DSP), use `residuez`.

    See Notes for details about the algorithm.

    Parameters
    ----------
    b : array_like
        Numerator polynomial coefficients.
    a : array_like
        Denominator polynomial coefficients.
    tol : float, optional
        The tolerance for two roots to be considered equal in terms of
        the distance between them. Default is 1e-3. See `unique_roots`
        for further details.
    rtype : {'avg', 'min', 'max'}, optional
        Method for computing a root to represent a group of identical roots.
        Default is 'avg'. See `unique_roots` for further details.

    Returns
    -------
    r : ndarray
        Residues corresponding to the poles. For repeated poles, the residues
        are ordered to correspond to ascending by power fractions.
    p : ndarray
        Poles ordered by magnitude in ascending order.
    k : ndarray
        Coefficients of the direct polynomial term.

    Warning
    -------
    This function may synchronize the device.

    See Also
    --------
    scipy.signal.residue
    invres, residuez, numpy.poly, unique_roots

    Notes
    -----
    The "deflation through subtraction" algorithm is used for
    computations --- method 6 in [1]_.

    The form of partial fraction expansion depends on poles multiplicity in
    the exact mathematical sense. However there is no way to exactly
    determine multiplicity of roots of a polynomial in numerical computing.
    Thus you should think of the result of `residue` with given `tol` as
    partial fraction expansion computed for the denominator composed of the
    computed poles with empirically determined multiplicity. The choice of
    `tol` can drastically change the result if there are close poles.

    References
    ----------
    .. [1] J. F. Mahoney, B. D. Sivazlian, "Partial fractions expansion: a
           review of computational methodology and efficiency", Journal of
           Computational and Applied Mathematics, Vol. 9, 1983.
    """
    if (cupy.issubdtype(b.dtype, cupy.complexfloating)
            or cupy.issubdtype(a.dtype, cupy.complexfloating)):
        b = b.astype(complex)
        a = a.astype(complex)
    else:
        b = b.astype(float)
        a = a.astype(float)

    b = cupy.trim_zeros(cupy.atleast_1d(b), 'f')
    a = cupy.trim_zeros(cupy.atleast_1d(a), 'f')

    if a.size == 0:
        raise ValueError("Denominator `a` is zero.")

    poles = roots(a)
    if b.size == 0:
        return cupy.zeros(poles.shape), _cmplx_sort(poles)[0], cupy.array([])

    if len(b) < len(a):
        k = cupy.empty(0)
    else:
        k, b = _polydiv(b, a)

    unique_poles, multiplicity = unique_roots(poles, tol=tol, rtype=rtype)
    unique_poles, order = _cmplx_sort(unique_poles)
    multiplicity = multiplicity[order]

    residues = _compute_residues(unique_poles, multiplicity, b)

    index = 0
    for pole, mult in zip(unique_poles, multiplicity):
        poles[index:index + mult] = pole
        index += mult

    return residues / a[0], poles, k


def residuez(b, a, tol=1e-3, rtype='avg'):
    """Compute partial-fraction expansion of b(z) / a(z).

    If `M` is the degree of numerator `b` and `N` the degree of denominator
    `a`::

                b(z)     b[0] + b[1] z**(-1) + ... + b[M] z**(-M)
        H(z) = ------ = ------------------------------------------
                a(z)     a[0] + a[1] z**(-1) + ... + a[N] z**(-N)

    then the partial-fraction expansion H(z) is defined as::

                 r[0]                   r[-1]
         = --------------- + ... + ---------------- + k[0] + k[1]z**(-1) ...
           (1-p[0]z**(-1))         (1-p[-1]z**(-1))

    If there are any repeated roots (closer than `tol`), then the partial
    fraction expansion has terms like::

             r[i]              r[i+1]                    r[i+n-1]
        -------------- + ------------------ + ... + ------------------
        (1-p[i]z**(-1))  (1-p[i]z**(-1))**2         (1-p[i]z**(-1))**n

    This function is used for polynomials in negative powers of z,
    such as digital filters in DSP.  For positive powers, use `residue`.

    See Notes of `residue` for details about the algorithm.

    Parameters
    ----------
    b : array_like
        Numerator polynomial coefficients.
    a : array_like
        Denominator polynomial coefficients.
    tol : float, optional
        The tolerance for two roots to be considered equal in terms of
        the distance between them. Default is 1e-3. See `unique_roots`
        for further details.
    rtype : {'avg', 'min', 'max'}, optional
        Method for computing a root to represent a group of identical roots.
        Default is 'avg'. See `unique_roots` for further details.

    Returns
    -------
    r : ndarray
        Residues corresponding to the poles. For repeated poles, the residues
        are ordered to correspond to ascending by power fractions.
    p : ndarray
        Poles ordered by magnitude in ascending order.
    k : ndarray
        Coefficients of the direct polynomial term.

    Warning
    -------
    This function may synchronize the device.

    See Also
    --------
    scipy.signal.residuez
    invresz, residue, unique_roots
    """

    if (cupy.issubdtype(b.dtype, cupy.complexfloating)
            or cupy.issubdtype(a.dtype, cupy.complexfloating)):
        b = b.astype(complex)
        a = a.astype(complex)
    else:
        b = b.astype(float)
        a = a.astype(float)

    b = cupy.trim_zeros(cupy.atleast_1d(b), 'b')
    a = cupy.trim_zeros(cupy.atleast_1d(a), 'b')

    if a.size == 0:
        raise ValueError("Denominator `a` is zero.")
    elif a[0] == 0:
        raise ValueError("First coefficient of determinant `a` must be "
                         "non-zero.")

    poles = roots(a)
    if b.size == 0:
        return cupy.zeros(poles.shape), _cmplx_sort(poles)[0], cupy.array([])

    b_rev = b[::-1]
    a_rev = a[::-1]

    if len(b_rev) < len(a_rev):
        k_rev = cupy.empty(0)
    else:
        k_rev, b_rev = _polydiv(b_rev, a_rev)

    unique_poles, multiplicity = unique_roots(poles, tol=tol, rtype=rtype)
    unique_poles, order = _cmplx_sort(unique_poles)
    multiplicity = multiplicity[order]

    residues = _compute_residues(1 / unique_poles, multiplicity, b_rev)

    index = 0
    powers = cupy.empty(len(residues), dtype=int)
    for pole, mult in zip(unique_poles, multiplicity):
        poles[index:index + mult] = pole
        powers[index:index + mult] = 1 + cupy.arange(int(mult))
        index += mult

    residues *= (-poles) ** powers / a_rev[0]

    return residues, poles, k_rev[::-1]



# distutils: language = c++

"""Thin wrapper of CUDA Driver API.

There are four differences compared to the original C API.

1. Not all functions are ported.
2. Errors are translated into CUDADriverError exceptions.
3. The 'cu' prefix of each API is omitted and the next character is set to
   lower case.
4. The resulting values are returned directly instead of references.

"""
cimport cython  # NOQA
from libc.stdint cimport intptr_t


###############################################################################
# Extern and Constants
###############################################################################

IF CUPY_USE_CUDA_PYTHON:
    from cuda.ccuda cimport *
    cdef inline void initialize():
        return
ELSE:
    include '_driver_extern.pxi'
    pass  # for cython-lint

cdef extern from '../../cupy_backend.h' nogil:
    # Build-time version
    # Note: CUDA_VERSION is defined either in CUDA Python or _driver_extern.pxi
    enum: HIP_VERSION

# Provide access to constants from Python.
from cupy_backends.cuda.api._driver_enum import *


###############################################################################
# Error handling
###############################################################################

class CUDADriverError(RuntimeError):

    def __init__(self, Result status):
        initialize()
        self.status = status
        cdef const char *name
        cdef const char *msg
        cuGetErrorName(status, &name)
        cuGetErrorString(status, &msg)
        cdef bytes s_name = name, s_msg = msg
        super(CUDADriverError, self).__init__(
            '%s: %s' % (s_name.decode(), s_msg.decode()))

    def __reduce__(self):
        return (type(self), (self.status,))


@cython.profile(False)
cpdef inline check_status(int status):
    if status != 0:
        raise CUDADriverError(status)


@cython.profile(False)
cdef inline void check_attribute_status(int status, int* pi) except *:
    # set attribute to -1 on older versions of CUDA where it was undefined
    if status == CUDA_ERROR_INVALID_VALUE:
        pi[0] = -1
    elif status != 0:
        raise CUDADriverError(status)


###############################################################################
# Build-time version
###############################################################################

cpdef get_build_version():
    """Returns the CUDA_VERSION / HIP_VERSION constant.

    Note that when built with CUDA Python support, CUDA_VERSION will become a
    constant:

    https://github.com/NVIDIA/cuda-python/blob/v11.4.0/cuda/ccuda.pxd#L2268

    In CuPy codebase, use `runtime.runtimeGetVersion()` instead of
    this function to change the behavior based on the target CUDA version.
    """

    # The versions are mutually exclusive
    if CUPY_CUDA_VERSION > 0:
        return CUDA_VERSION
    elif CUPY_HIP_VERSION > 0:
        return HIP_VERSION
    else:
        return 0


cpdef bint _is_cuda_python():
    return CUPY_USE_CUDA_PYTHON


###############################################################################
# Primary context management
###############################################################################

cpdef devicePrimaryCtxRelease(Device dev):
    initialize()
    with nogil:
        status = cuDevicePrimaryCtxRelease(dev)
    check_status(status)

###############################################################################
# Context management
###############################################################################

cpdef intptr_t ctxGetCurrent() except? 0:
    initialize()
    cdef Context ctx
    with nogil:
        status = cuCtxGetCurrent(&ctx)
    check_status(status)
    return <intptr_t>ctx

cpdef ctxSetCurrent(intptr_t ctx):
    initialize()
    with nogil:
        status = cuCtxSetCurrent(<Context>ctx)
    check_status(status)

cpdef intptr_t ctxCreate(Device dev) except? 0:
    initialize()
    cdef Context ctx
    cdef unsigned int flags = 0
    with nogil:
        status = cuCtxCreate(&ctx, flags, dev)
    check_status(status)
    return <intptr_t>ctx

cpdef ctxDestroy(intptr_t ctx):
    initialize()
    with nogil:
        status = cuCtxDestroy(<Context>ctx)
    check_status(status)

cpdef int ctxGetDevice() except? -1:
    initialize()
    cdef Device dev
    with nogil:
        status = cuCtxGetDevice(&dev)
    check_status(status)
    return dev


###############################################################################
# Module load and kernel execution
###############################################################################

cpdef intptr_t linkCreate() except? 0:
    initialize()
    cdef LinkState state
    with nogil:
        status = cuLinkCreate(0, <CUjit_option*>0, <void**>0, &state)
    check_status(status)
    return <intptr_t>state


cpdef linkAddData(intptr_t state, int input_type, bytes data, unicode name):
    initialize()
    cdef const char* data_ptr = data
    cdef size_t data_size = len(data) + 1
    cdef bytes b_name = name.encode()
    cdef const char* b_name_ptr = b_name
    with nogil:
        status = cuLinkAddData(
            <LinkState>state, <CUjitInputType>input_type, <void*>data_ptr,
            data_size, b_name_ptr, 0, <CUjit_option*>0, <void**>0)
    check_status(status)


cpdef linkAddFile(intptr_t state, int input_type, unicode path):
    initialize()
    cdef bytes b_path = path.encode()
    cdef const char* b_path_ptr = b_path
    with nogil:
        status = cuLinkAddFile(<LinkState>state, <CUjitInputType>input_type,
                               b_path_ptr, 0, <CUjit_option*>0, <void**>0)
    check_status(status)


cpdef bytes linkComplete(intptr_t state):
    initialize()
    cdef void* cubinOut
    cdef size_t sizeOut
    with nogil:
        status = cuLinkComplete(<LinkState>state, &cubinOut, &sizeOut)
    check_status(status)
    return bytes((<char*>cubinOut)[:sizeOut])


cpdef linkDestroy(intptr_t state):
    initialize()
    with nogil:
        status = cuLinkDestroy(<LinkState>state)
    check_status(status)


cpdef intptr_t moduleLoad(str filename) except? 0:
    initialize()
    cdef Module module
    cdef bytes b_filename = filename.encode()
    cdef char* b_filename_ptr = b_filename
    with nogil:
        status = cuModuleLoad(&module, b_filename_ptr)
    check_status(status)
    return <intptr_t>module


cpdef intptr_t moduleLoadData(bytes image) except? 0:
    initialize()
    cdef Module module
    cdef char* image_ptr = image
    with nogil:
        status = cuModuleLoadData(&module, image_ptr)
    check_status(status)
    return <intptr_t>module


cpdef moduleUnload(intptr_t module):
    initialize()
    with nogil:
        status = cuModuleUnload(<Module>module)
    check_status(status)


cpdef intptr_t moduleGetFunction(intptr_t module, str funcname) except? 0:
    initialize()
    cdef Function func
    cdef bytes b_funcname = funcname.encode()
    cdef char* b_funcname_ptr = b_funcname
    with nogil:
        status = cuModuleGetFunction(&func, <Module>module, b_funcname_ptr)
    check_status(status)
    return <intptr_t>func


cpdef intptr_t moduleGetGlobal(intptr_t module, str varname) except? 0:
    initialize()
    cdef Deviceptr var
    cdef size_t size
    cdef bytes b_varname = varname.encode()
    cdef char* b_varname_ptr = b_varname
    with nogil:
        status = cuModuleGetGlobal(&var, &size, <Module>module, b_varname_ptr)
    check_status(status)
    return <intptr_t>var


cpdef launchKernel(
        intptr_t f, unsigned int grid_dim_x, unsigned int grid_dim_y,
        unsigned int grid_dim_z, unsigned int block_dim_x,
        unsigned int block_dim_y, unsigned int block_dim_z,
        unsigned int shared_mem_bytes, intptr_t stream, intptr_t kernel_params,
        intptr_t extra):
    initialize()
    with nogil:
        status = cuLaunchKernel(
            <Function>f, grid_dim_x, grid_dim_y, grid_dim_z,
            block_dim_x, block_dim_y, block_dim_z,
            shared_mem_bytes, <Stream>stream,
            <void**>kernel_params, <void**>extra)
    check_status(status)


cpdef launchCooperativeKernel(
        intptr_t f, unsigned int grid_dim_x, unsigned int grid_dim_y,
        unsigned int grid_dim_z, unsigned int block_dim_x,
        unsigned int block_dim_y, unsigned int block_dim_z,
        unsigned int shared_mem_bytes, intptr_t stream,
        intptr_t kernel_params):
    initialize()
    with nogil:
        status = cuLaunchCooperativeKernel(
            <Function>f, grid_dim_x, grid_dim_y, grid_dim_z,
            block_dim_x, block_dim_y, block_dim_z,
            shared_mem_bytes, <Stream>stream, <void**>kernel_params)
    check_status(status)


###############################################################################
# Function attributes
###############################################################################

# -1 is reserved by check_attribute_status
cpdef int funcGetAttribute(int attribute, intptr_t f) except? -2:
    initialize()
    cdef int pi
    with nogil:
        status = cuFuncGetAttribute(
            &pi,
            <CUfunction_attribute> attribute,
            <Function> f)
    check_attribute_status(status, &pi)
    return pi


cpdef funcSetAttribute(intptr_t f, int attribute, int value):
    initialize()
    with nogil:
        status = cuFuncSetAttribute(
            <Function> f,
            <CUfunction_attribute> attribute,
            value)
    check_status(status)


###############################################################################
# Occupancy
###############################################################################

cpdef int occupancyMaxActiveBlocksPerMultiprocessor(
        intptr_t func, int blockSize, size_t dynamicSMemSize):
    initialize()
    cdef int numBlocks
    with nogil:
        status = cuOccupancyMaxActiveBlocksPerMultiprocessor(
            &numBlocks, <Function> func, blockSize, dynamicSMemSize)
    check_status(status)
    return numBlocks


cpdef occupancyMaxPotentialBlockSize(intptr_t func, size_t dynamicSMemSize,
                                     int blockSizeLimit):
    initialize()
    # CUoccupancyB2DSize is set to NULL as there is no way to pass in a
    # unary function from Python.
    cdef int minGridSize, blockSize
    with nogil:
        status = cuOccupancyMaxPotentialBlockSize(
            &minGridSize, &blockSize, <Function> func, <CUoccupancyB2DSize>
            NULL, dynamicSMemSize, blockSizeLimit)
    check_status(status)
    return minGridSize, blockSize


###############################################################################
# Stream management
###############################################################################

cpdef intptr_t streamGetCtx(intptr_t stream) except? 0:
    initialize()
    cdef Context ctx
    with nogil:
        status = cuStreamGetCtx(<Stream>stream, &ctx)
    check_status(status)
    return <intptr_t>ctx



import os as _os
import threading as _threading

from cupy_backends.cuda.api cimport runtime


cdef object _thread_local = _threading.local()


cdef bint _ptds = bool(int(
    _os.environ.get('CUPY_CUDA_PER_THREAD_DEFAULT_STREAM', '0')) != 0)


cdef class _ThreadLocal:
    cdef list current_stream  # list of intptr_t

    def __init__(self):
        cdef int i, num_devices = runtime.getDeviceCount()
        self.current_stream = [0 for i in range(num_devices)]

    @staticmethod
    cdef _ThreadLocal get():
        try:
            tls = _thread_local.tls
        except AttributeError:
            tls = _thread_local.tls = _ThreadLocal()
        return <_ThreadLocal>tls

    cdef set_current_stream_ptr(self, intptr_t ptr, int device_id=-1):
        if device_id == -1:
            device_id = runtime.getDevice()
        self.current_stream[device_id] = ptr

    cdef intptr_t get_current_stream_ptr(self, int device_id=-1):
        # Returns the stream previously set, otherwise returns
        # nullptr or runtime.streamPerThread when
        # CUPY_CUDA_PER_THREAD_DEFAULT_STREAM=1.
        if device_id == -1:
            device_id = runtime.getDevice()
        cdef intptr_t curr_stream = self.current_stream[device_id]
        if curr_stream == 0 and is_ptds_enabled():
            return runtime.streamPerThread
        return curr_stream


cdef intptr_t get_current_stream_ptr():
    """C API to get current CUDA stream pointer.

    Returns:
        intptr_t: The current CUDA stream pointer.
    """
    tls = _ThreadLocal.get()
    return <intptr_t>tls.get_current_stream_ptr()


cdef set_current_stream_ptr(intptr_t ptr, int device_id=-1):
    """C API to set current CUDA stream pointer.

    Args:
        ptr (intptr_t): CUDA stream pointer.
        device_id (int): device ID. Look up the current device if -1.

    .. warning::

        This method is intended to be called from `cupy.cuda.stream` module.
        Do not call this method from somewhere else; this method only changes
        the default stream for `cupy_backends.*`, so the stream used will be
        inconsistent with the default one for `cupy.*`.

    """
    tls = _ThreadLocal.get()
    tls.set_current_stream_ptr(ptr, device_id)


# cpdef for unit testing
cpdef intptr_t get_default_stream_ptr():
    """Get the CUDA default stream pointer.

    Returns:
        intptr_t: CUDA stream pointer.
    """
    if is_ptds_enabled():
        return runtime.streamPerThread
    else:  # we don't return 0 here
        return runtime.streamLegacy


cdef bint is_ptds_enabled():
    if runtime._is_hip_environment:
        # HIP does not support PTDS, just ignore the env var
        return False
    return _ptds
